(* ::Package:: *)

(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Provers`RewriteInteractiveProver`"]

Needs[ "Theorema`Provers`"]
Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]

(* Remark: The names of the inference rules in this prover carry the prefix "RRP", because they originate from
	the ReductionRingProver. Changing names of inference rules _after_ proofs have already been carried out
	is not so easy. Still, every single inference rule defined below is absolutely general and does not depend
	on any "implicit" knowledge about notions that are not built-in Theorema notions. *)
	
With[ {lang = "English"},
	MessageName[ $hoLoaded, "usage", lang] = "Indicates whether the optional higher-order rewriting package has been loaded successfully.";
	MessageName[ $ruleGenerationOptions, "usage", lang] = "This global variable specifies how rewrite rules are generated.";
	MessageName[ $extractTagsForRewriting, "usage", lang] = "This variable specifies whether tags of formulas (of the form \"[...]\") should be taken into account when generating rewrite rules.";
	MessageName[ $backwardGround, "usage", lang] = "This variable specifies which kind of backward rules shall be used for condition-checking in all sorts of inference rules. This does *not* include goal rewriting.";
	MessageName[ $proposedRules, "usage", lang] = "";
	MessageName[ getRulesFromRW, "usage", lang] = "";
	MessageName[ getRuleClassesFromRW, "usage", lang] = "";
	MessageName[ followsImmediately, "usage", lang] = "";
	MessageName[ alphaMemberQ, "usage", lang] = "";
	MessageName[ addNewKB, "usage", lang] = "";
	MessageName[ joinNewKB, "usage", lang] = "";
	MessageName[ RRPToBeProved, "usage", lang] = "";
	MessageName[ makeRRPFML, "usage", lang] = "";
	MessageName[ activatedQ, "usage", lang] = "";
	MessageName[ goalActivated, "usage", lang] = "";
	MessageName[ preprocessExpr, "usage", lang] = "Prepares the given expression for rewriting.";
	MessageName[ postprocessExpr, "usage", lang] = "Turns a given result of rewriting back into a well-formed Theorema expression.";
	MessageName[ genGroundRule, "usage", lang] = "";
];

Quiet[ Check[ Get[ "Theorema`Language`HORewriting`"]; $hoLoaded = True, $hoLoaded = False]];

Begin["`Private`"]

(* TODO:
? Add inference rule for making implicit/hidden knowledge explicit, i.e. adding formulas only available as rewrite rules
	to the knowledge base.
? Allow formula references in 'Introduce Arbitrary Formula', meaning that e.g. every identifier of the form "?G#1"
	is replaced by the formula with label "G#1", given that it appears in the current proof situation.
? Store rewrite rules originating from [IR]-formulas in *global* variable rather than in each proof situation.
	Attention! This global variable must be saved in file as well, for resuming with proofs.
? Treat existentially quantified formulas in 'followsImmediately' by matching.
- The computational simplification of "Let"-expressions might be dangerous: no capture-avoiding substitution, but mere replacement.
*)



(* ::Section::Closed:: *)
(* Auxiliary Functions *)


(* ::Subsection::Closed:: *)
(* Dialogs *)


Options[ RRPDialogInput] = {"CancelTooltip" :> translate[ "cancel"], "OKTooltip" :> translate[ "accept"],
	"CloseOnEscape" -> True, "AdditionalMenuItems" -> {}, "Docked" -> True};
RRPDialogInput[ Notebook[ data_List, ___], Hold[ okAction_], opts___?OptionQ] :=
	Module[ {d, dc, menu, cancelTooltip, okTooltip, escAction, additionalMenu, docked},
		{cancelTooltip, okTooltip, additionalMenu, escAction, docked} =
			{"CancelTooltip", "OKTooltip", "AdditionalMenuItems", "CloseOnEscape", "Docked"} /. {opts} /. Options[ RRPDialogInput];
		escAction = If[ escAction,
						"EscapeKeyDown" :> DialogReturn[ $Canceled],
					(*else*)
						"EscapeKeyDown" :> Paste[ "\[AliasDelimiter]"]
					];
		menu = Cell[ BoxData[ ToBoxes[ Row[ Join[ {
						Tooltip[ Button[ "<<", DialogReturn[ $Canceled], Appearance -> "Frameless"], cancelTooltip],
						Tooltip[ Button[ "OK", DialogReturn[ okAction], ImageSize -> 70, Appearance -> "Frameless"], okTooltip]
					}, additionalMenu], Spacer[ 10]
				]]], "Hint", Background -> TMAcolor[ 1]];

		{d, dc} = If[ TrueQ[ docked],
						{data, menu},
						{Prepend[ data, menu], {}}
					];

		DialogInput[
			Notebook[
				d,
				DockedCells -> dc,
				StyleDefinitions -> makeColoredStylesheet[ "Dialog"],
				Magnification -> CurrentValue[ First[ getTheoremaCommander[]], Magnification],
				Editable -> False,
				ShowCellBracket -> False,
				WindowSize -> All,
				WindowElements -> {"VerticalScrollBar", "HorizontalScrollBar", "StatusArea"},
				Sequence @@ DeleteCases[ {opts}, (Rule|RuleDelayed)[ _String?(MemberQ[ Options[ RRPDialogInput], _[ #, _]] &), _]]
			],
			NotebookEventActions -> {escAction, "ReturnKeyDown" :> DialogReturn[ okAction], "WindowClose" :> DialogReturn[ $Canceled]}
		]
	]

kbBrowserNB[] :=
	Module[ {nbFiles, view, allNB, addGlobal = False, kbBrowseSelection = $tcKBBrowseSelection["prove"], out},
		If[ $tmaNbUpdateQueue =!= {},
			(* We don't use 'refreshKBStruct', because we want to update all notebooks (regardless of their timestamp),
				and we don't want to see any messages concerning notebook-integrity violations. *)
			nbFiles = $tmaNbUpdateQueue[[All, 1]];
			Scan[ updateKBBrowser, nbFiles]
		];
		If[ $kbStruct =!= {} && ValueQ[ kbBrowseSelection],
			allNB = Map[ First, $kbStruct];
			Scan[ With[ {key = First[ #]}, out[ key] = False] &, $tmaEnv];
			view = {
				Cell[ BoxData[ ToBoxes[
						If[ Plus@@Map[ StringLength[ FileBaseName[ #]]&, allNB] > 50, 
							PopupMenu[ Dynamic[ kbBrowseSelection], 
								Map[ First[ #] -> Tooltip[ FileBaseName[ First[ #]], First[ #]]&, $kbStruct]
							], 
							Dynamic[ Refresh[ Row[ Map[
								Button[ Tooltip[ Style[ FileBaseName[ #], FontColor -> If[ kbBrowseSelection === #, TMAcolor[ 4], TMAcolor[ 13]]], #],
									kbBrowseSelection = #, 
									Background -> If[ kbBrowseSelection === #, TMAcolor[ 6], TMAcolor[ 0]],
									FrameMargins -> 2, ImageMargins -> 0, ImageSize -> Automatic
								]&,
								allNB
							]], TrackedSymbols :> {kbBrowseSelection}]]
						]
				]]],
				Cell[ BoxData[ToBoxes[ Dynamic[ Refresh[
						Pane[ structView[ kbBrowseSelection, Replace[ kbBrowseSelection, $kbStruct], out], ImageSize -> {350, Automatic}], 
           				TrackedSymbols :> {kbBrowseSelection}
           			]]]],
           			Selectable -> False
           		]
			};
			
			RRPDialogInput[
				Notebook[ view],
				Hold[ Append[ Cases[ $tmaEnv, FML$[ key_ /; out[ key], ___]], addGlobal]],
				"AdditionalMenuItems" -> {Tooltip[
											Toggler[ Dynamic[ addGlobal], {True -> Text[ translate[ "addGlob"]], False -> Text[ translate[ "addLoc"]]}, Appearance -> "Frameless"],
											translate[ "ttAddGlobLoc"]
										]},
				WindowTitle -> translate[ "addKnowledge"]
			],
	    (*else*)
			$Failed
		]
	]

structView[ file_, cells_, fun_] :=
	With[ {view = structView[ file, cells, {}, {}, fun]},
		If[ view === {},
			translate[ "noKnowl"],
		(*else*)
			First[view]
		]
	]
structView[ file_, {head:Cell[ sec_, style:("Title"|"Section"|"Subsection"|"Subsubsection"|"Subsubsubsection"|"EnvironmentHeader"), opts___], rest___}, tags_, headers_List, fun_] :=
	With[ {content = DeleteCases[ Map[ structView[ file, #, tags, Append[ headers, sec], fun]&, {rest}], {}]},
		If[ content === {},
			{},
		(*else*)
			Module[ {views, compTags},
				{views, compTags} = Transpose[ content];
				compTags = Union@@compTags;
				{
					OpenerView[
						{headerView[ file, head, compTags, fun], Column[views]},
						MemberQ[ {"Title", "Section", "Subsection"}, style]
					],
					compTags
				}
			]
		]
	]
structView[ file_, item_List, tags_, headers_List, fun_] :=
	With[ {content = DeleteCases[ Map[ structView[ file, #, tags, headers, fun]&, item], {}]},
		If[ content === {},
			{},
		(*else*)
			Module[ {views, compTags},
				{views, compTags} = Transpose[ content];
				compTags = Union@@compTags;
				{Column[ views], compTags}
			]
		]
	]
structView[ file_, Cell[content_, "FormalTextInputFormula", a___, CellTags -> cellTags_, b___], tags_, headers_List, fun_] :=
	With[ {cleanCellTags = getCleanCellTags[ cellTags], keyTags = getKeyTags[ cellTags]},
	With[ {formPos = Position[ {$tmaEnv, $tmaArch}, FML$[ keyTags, _, __], {2}, 1]},
		(*check whether cell has been evaluated -> formula is in KB?*)
		If[ formPos =!= {},
			With[ {nbAvail = FileExistsQ[ file] && FileExtension[ file] === "nb"},
			Module[ {formulaLabel, idLabel},
				(*Join list of CellTags, use $labelSeparator.*)
				If[ cleanCellTags === {},
					formulaLabel = cellTagsToString[ cellTags];
					idLabel = formulaLabel,
				(*else*)
					formulaLabel = cellTagsToString[ cleanCellTags];
					idLabel = getCellIDLabel[ keyTags]
				];
				{Row[ {
						Which[ kbSelectProve[ keyTags],
							"\[Checkmark]",
							
							MatchQ[ Theorema`Interface`GUI`Private`$selectedProofGoal, FML$[ keyTags, __]],
							"\[Times]",
							
			     			True,
							Checkbox[ Dynamic[ fun["KEY"]]] /. "KEY" -> keyTags
						],
						Tooltip[ Hyperlink[ Style[ formulaLabel, "FormalTextInputFormula"],
								{file, idLabel}, Enabled -> nbAvail, ActiveStyle -> If[ nbAvail, "HyperlinkActive", None]],
							theoremaDisplay[ Extract[ {$tmaEnv, $tmaArch}, Append[ First[ formPos], 2]]]]
					}, Spacer[ 10]
				],
				{keyTags}}
			]],
    	(*else*)
			{}
		]
	]]
structView[ file_, Cell[ content_, "FormalTextInputFormula", ___], tags_, headers_List, _] := {}

headerView[ file_, Cell[ content_String, ___], tags_, fun_] :=
	Row[
		{
			Checkbox[ Dynamic[ allTrue[ tags, fun], setAll[ tags, fun, #]&]],
			Text[ StringReplace[ content, "\n"|"\t" -> " "]]
		},
		Spacer[ 10]
	]
headerView[ file_, Cell[ content_TextData, style_, ___], tags_, fun_] :=
	headerView[ file, Cell[ Theorema`Interface`GUI`Private`formattedCellToString[ content], style], tags, fun]

(* 'RRPBoxesToExpression' returns its results inside 'Expression', even if they are 'Null' or '$Failed'. *)
RRPBoxesToExpression[ l_List] :=
	Module[ {out},
		With[ {lnew = Replace[ removeRedundantBoxes[ l], RowBox[ {"(", expr___, ")"}] :> RowBox[ {"Theorema`Language`SEQ$", "[", expr, "]"}], {1}]
						//. {FormBox[ box_, TheoremaForm] :> box, TagBox[ box_, DisplayForm] :> box, TagBox[ InterpretationBox[ _, box_], "Boxes"] :> box}},
			Block[ {$parseTheoremaExpressions = True, $ContextPath = Join[ {"Theorema`Language`"}, $TheoremaArchives, $ContextPath]},
				out = Replace[ lnew, term:Except[ Null] :> Quiet[ Check[ ToExpression[ term, StandardForm, Hold], $Failed]], {1}];
				Begin[ "Theorema`Knowledge`"];
				out = Replace[ out, {e_Hold :> Replace[ makeTmaExpression[ e], {SEQ$[ x_] :> Expression[ x], x_ :> Expression[ x]}], e_ :> Expression[ e]}, {1}];
				End[]
			]
		];
		out
	]
RRPBoxesToExpression[ b_] := First[ RRPBoxesToExpression[ {b}]]
	
instantiationDialog[ v_, const_List, {g_, kb_}, k_, opts___?OptionQ] :=
	Module[ {expr, fixBut},
		expr[ _] = Null;
		fixBut = Map[ ToBoxes,
					Replace[
						const,
						{
							SIMPRNG$[ a_Symbol] :> PasteButton[ theoremaDisplay[ a]],
							SIMPRNG$[ a_?AtomQ] :> PasteButton[ DisplayForm[ a], a],
							r:(_[ c_, ___]) :> PasteButton[ theoremaDisplay[ RNG$[ r]],
											With[ {fbox = theoremaBoxes[ c]},
												DisplayForm[ InterpretationBox[ fbox, c]]
											]
										  ]
						},
						{1}
					]
				];
		
		RRPDialogInput[ Notebook[ {
								Cell[ BoxData[ GridBox[
									MapIndexed[ {ToBoxes[ theoremaDisplay[ RNG$[ #1]]], ":=", 
										ToBoxes[ InputField[ Dynamic[ expr[ #2[[1]]]], Boxes, FieldSize -> Medium]]}&, v]
									]], "Text", Deployed -> True
								],
								Cell[ BoxData[ RowBox[ Join[ {
										ToBoxes[ Text[ translate[ "availConst"] <> ":"]], 
										ToBoxes[ Spacer[5]]}, fixBut]
									]], "Text", ShowStringCharacters -> False, Deployed -> True],
								Cell[ TextData[ translate[ "proofSituation"]], "Subsubsection"],
								makeFormulaHighlightCells[ g, kb, FML$[ k, __], FML$[ k, __]]}
						],
				Hold[ Array[ expr, Length[ v]]],
				"CloseOnEscape" -> False,
				opts
			]
	]

instantiateForallKBInteractive[ _List, RNG$[], _List, _] := $Failed
instantiateForallKBInteractive[ const_List, rng_RNG$, {g_FML$, kb_List}, k_] :=
	Module[{checkConds = True, keepOrig = True},
	With[ {inst = instantiationDialog[ List @@ rng, const, {g, kb}, k,
					"AdditionalMenuItems" -> {
						Tooltip[ Row[ {Checkbox[ Dynamic[ checkConds]], Spacer[ 5], Text[ translate[ "proveDomCondsImmediately"]]}], translate[ "ttProveDomCondsImmediately"]],
						Tooltip[ Row[ {Checkbox[ Dynamic[ keepOrig]], Spacer[ 5], Text[ translate[ "keepOrig"]]}], translate[ "ttKeepRemoveOrig"]]
					},
					WindowTitle -> translate[ "instantiateForall"]
				]},
		If[ ListQ[ inst],
			Join[ RRPBoxesToExpression[ inst], {checkConds, keepOrig}],
		(*else*)
			$Failed
		]
	]]

instantiateExistsGoalInteractive[ g:FML$[ k_, Theorema`Language`Exists$TM[ rng_, __], ___], const_List, kb_List] :=
	With[ {inst = instantiationDialog[ List @@ rng, const, {g, kb}, k, WindowTitle -> translate[ "instantiateExists"]]},
		If[ inst === $Failed,
			$Failed,
		(*else*)
			RRPBoxesToExpression[ inst]
		]
	]

makeSubstitution[ vars_List, exprs_List] :=
	If[ Length[ exprs] === Length[ vars],
		DeleteCases[
			MapThread[
				Switch[ #1,
					Null|"_",
					$Failed,
					
					Hold[ SEQ$[]|(VAR$|FIX$|META$)[ _SEQ0$, ___]],
					If[ MatchQ[ #2, _[ _SEQ0$]],
						#2 -> #1,
					(*else*)
						$Failed
					],
					
					Hold[ SEQ$[ _, __]],
					Switch[ #2,
						_[ _SEQ1$],
						If[ MemberQ[ #1, Except[ (VAR$|FIX$|META$)[ _SEQ0$, ___]], {2}],
							#2 -> #1,
						(*else*)
							$Failed
						],
						_[ _SEQ0$],
						#2 -> #1,
						_,
						$Failed
					],
					
					Hold[ (VAR$|FIX$|META$)[ _SEQ1$, ___]],
					If[ MatchQ[ #2, _[ _SEQ0$|_SEQ1$]],
						#2 -> #1,
					(*else*)
						$Failed
					],
					
					_,
					#2 -> #1
				]&,
				{Replace[ exprs, Hold[ SEQ$[ expr_]] :> Hold[ expr], {1}], vars}
			],
			$Failed
		],
	(*else*)
		$Failed
	]

selectFormulaDialog[ goal_FML$, k_List, title_String, opts___?OptionQ] :=
	selectFormulaDialog[ goal, k, Null, FML$[ BlankNullSequence[]], title, opts]
selectFormulaDialog[ goal_FML$, k_List, pattern:Except[ _FML$], title_String, opts___?OptionQ] :=
	selectFormulaDialog[ goal, k, Null, FML$[ Blank[], pattern, BlankNullSequence[]], title, opts]
selectFormulaDialog[ goal_FML$, k:{___FML$}, gPattern_, kPattern_, title_String, opts___?OptionQ] :=
	Module[ {selectedFormula = 0, cells, index},
		{cells, index} = makeFormulaChoiceCells[ goal, k, gPattern, kPattern, Hold[ selectedFormula]];
		Which[
			index === {},
			$Failed,
			
			Length[ index] === 1 && !TrueQ[ "DialogOneOption" /. {opts} /. "DialogOneOption" -> False],
			First[ index] - 1,
			
			True,
			RRPDialogInput[ Notebook[ {cells}],
				Hold[ selectedFormula - 1],
				WindowTitle -> translate[ title],
				opts
			]
		]
	]

makeFormulaChoiceCells[ g_FML$, kb_List, gPattern_, kPattern_, out_Hold] :=
	makeFormulaChoiceCells[ {g}, kb, gPattern, kPattern, out]
makeFormulaChoiceCells[ gl:{___FML$}, kb:{___FML$}, gPattern_, kPattern_, Hold[ out_]] :=
	With[ {l = Length[ gl]}, 
		Module[ {ac, lab, ret, j = {}},
			ret = MapIndexed[ Function[ {g, index}, 
							With[ {i = First[ index]},
								lab = makeLabel[ label@g];
								{Cell[ BoxData[
									If[ MatchQ[ g, gPattern],
										AppendTo[ j, i];
										ToBoxes[ RadioButton[ Dynamic[ out], i]],
										RowBox[{}]
									]], Deployed -> True],
								Cell[ BoxData[ selectableExpressionBox[ g]], "Goal"],
								Cell[ lab, "GoalLabel", Deployed -> True]}
							]
						],
						gl
					];
			ac = MapIndexed[ Function[ {k, index}, 
							With[ {i = First[ index]},
								lab = makeLabel[ label@k];
								{Cell[ BoxData[
									If[ MatchQ[ k, kPattern],
										AppendTo[ j, i + l];
										ToBoxes[ RadioButton[ Dynamic[ out], i + l]],
										RowBox[ {}]
									]], Deployed -> True],
								Cell[ BoxData[ selectableExpressionBox[ k]], "Assumption"],
								Cell[ lab, "AssumptionLabel", Deployed -> True]}
							]
						],
						kb
					];
			If[ ac =!= {},
				ret = Join[ ret, ac]
			];
			If[ ret =!= {}, ret = GridBox[ ret, ColumnAlignments -> Left, RowLines -> {True, False}]];
			{ret, j}
		]
	]

makeFormulaHighlightCells[ g_FML$, kb_List, gPattern_, kPattern_] :=
	makeFormulaHighlightCells[ {g}, kb, gPattern, kPattern]
makeFormulaHighlightCells[ gl:{___FML$}, kb:{___FML$}, gPattern_, kPattern_] :=
	Module[ {ac, lab, ret},
		ret = Map[ Function[ g, 
						lab = makeLabel[ label@g];
						{Cell[ TextData[
							If[ MatchQ[ g, gPattern],
								"\[FilledRightTriangle]",
								""
							]], Deployed -> True],
						Cell[ BoxData[ selectableExpressionBox[ g]], "Goal"],
						Cell[ lab, "GoalLabel", Deployed -> True]}
					],
					gl
				];
		ac = Map[ Function[ k, 
						lab = makeLabel[ label@k];
						{Cell[ TextData[
							If[ MatchQ[ k, kPattern],
								"\[FilledRightTriangle]",
								""
							]], Deployed -> True],
						Cell[ BoxData[ selectableExpressionBox[ k]], "Assumption"],
						Cell[ lab, "AssumptionLabel", Deployed -> True]}
					],
					kb
				];
		If[ ac =!= {}, 
			ret = Join[ ret, ac]
		];
		If[ ret =!= {}, ret = GridBox[ ret, ColumnAlignments -> Left, RowLines -> {True, False}]];
		ret
	]

selectGoalDialog[ goal:{__FML$}, k:{__FML$}, title_String, opts___?OptionQ] :=
	DynamicModule[ {selectedFormula = 0}, 
		RRPDialogInput[ Notebook[ {First[ makeFormulaChoiceCells[ goal, k, Blank[], Null, Hold[ selectedFormula]]]}],
			Hold[ selectedFormula],
			WindowTitle -> translate[ title],
			opts
		]
	]
	
rewritingDialog[ g_FML$, kb_List, rw_List, ruleTypes_List, Hold[ fmlActivated_]] :=
	Module[ {rules = Cases[ rw, HoldPattern[ type_String?(MemberQ[ ruleTypes, # -> _]&) -> _List]],
			forms, ruleTags, ruleForms, ruleActivated, fmlTags, ps, mode = 1, cc = True, pc = True, keepOriginal = False, t = 1, gk = key[g]},
		rules = Column[ DeleteCases[ Replace[ rules, (type_String -> l_List) :>
								If[ l === {},
									$Failed,
								(*else*)
									forms = DeleteDuplicates[ Map[ extractFormulaFromRule, l]];
									ruleForms[ type] = forms;
									ruleTags[ type] = Map[ key, forms];
									ruleActivated[ type][_] = False;
									OpenerView[ {
										Row[ {
											Checkbox[ Dynamic[
												allTrue[ ruleTags[ type], ruleActivated[ type]], 
												setAll[ ruleTags[ type], ruleActivated[ type], #]&]
											], 
											Text[ Style[ Replace[ type, ruleTypes], Bold]]}, Spacer[ 5]
										], 
										Column[ Map[ With[ {k = key[#]}, 
											Row[ {Checkbox[ Dynamic[ ruleActivated[ type][ k]]], 
												Tooltip[ theoremaDisplay[ formula[ #]], label[ #]]}, Spacer[ 5]]]&, forms]
										]},
										True
									]
								],
							{1}
						],
						$Failed
					]
				];
				
		fmlActivated[ gk] = goalActivated[];
		fmlTags = Prepend[ Map[ With[ {k = key[ #]}, fmlActivated[ k] = activatedQ[ k]; k]&, kb], gk];
		ps = Column[ Join[ {
						Row[ {
							Checkbox[ Dynamic[
								allTrue[ fmlTags, fmlActivated],
								setAll[ fmlTags, fmlActivated, #]&]
							], 
							Text[ Style[ translate[ "proofSituation"], Bold]]}, Spacer[ 5]
						], 
						(*Row[ {Spacer[ 25], Text[ translate[ "goal"]]}],*)
						Row[ {Checkbox[ Dynamic[ fmlActivated[ gk]]], Tooltip[ theoremaDisplay[ formula@g], label@g]}, Spacer[ 5]]
						(*,Row[ {Spacer[ 25], Text[ translate[ "knowledge"]]}]*)
					}, 
					Map[ With[ {k = key[ #]}, 
						Row[ {Checkbox[ Dynamic[ fmlActivated[k]]], Tooltip[ theoremaDisplay[ formula[#]], label[ #]]}, Spacer[5]]
						]&, kb
					]
				],
				Dividers -> {None, {False, False, True, False}}
			];
  
		RRPDialogInput[ Notebook[ {
				Cell[ BoxData[ ToBoxes[
					Column[ {
						Row[ {Checkbox[ Dynamic[ cc]], Tooltip[ Text[ translate[ "checkConds"]], translate[ "ttCheckConds"]]}, Spacer[ 5]],
						Row[ {Checkbox[ Dynamic[ pc]], Tooltip[ Text[ translate[ "proveConds"]], translate[ "ttProveConds"]]}, Spacer[ 5]],
						Row[ {Checkbox[ Dynamic[ keepOriginal]], Tooltip[ Text[ translate[ "keepOrig"]], translate[ "ttKeepRemoveOrig"]]}, Spacer[ 5]]
					}]]],
					ShowStringCharacters -> False,
					Deployed -> True
				],
				Cell[ BoxData[ ToBoxes[
					Tooltip[
						PopupView[ {translate[ "scanRulesFirst"], translate[ "scanRulesAll"], translate[ "scanRulesRepeated"], translate[ "scanFormulas"]}, Dynamic[mode]],
						translate[ "ttRewritingControl"]
					]]], 
					Deployed -> True
				], 
				Cell[ BoxData[ ToBoxes[ TabView[ {translate[ "rewriteRules"] -> rules, translate[ "proofSituation"] -> ps}, Dynamic[ t]]]]]
			}], 
			Hold[ {Map[ With[ {tp = First[ #]}, Join @@ Cases[ rw, (tp -> l_List) :> Select[ l, Composition[ ruleActivated[ tp], First]]]]&, ruleTypes], mode, cc, pc, keepOriginal}],
			WindowTitle -> translate[ "rewriteProofSit"]
		]
	]
	
formulaInputDialog[ g_, kb_List, const_List, opts___?OptionQ] :=
	Module[ {caseDistinction = False, performReplacements = False, expr = Null, menu,
				fixBut = Map[ ToBoxes[ PasteButton[ theoremaDisplay[ RNG$[ #]], 
							With[ {fbox = ToBoxes[ First[ #], TheoremaForm], fc = First[ #]}, 
								DisplayForm[ InterpretationBox[ fbox, fc]]]]
							]&, const]},
		menu = {Tooltip[ Toggler[ Dynamic[ caseDistinction], {True -> Text[ translate[ "distinguishCases"]], False -> Text[ translate[ "proveFormula"]]}], 
						translate[ "ttDistinguishProve"]
				], 
				Tooltip[ Row[ {Checkbox[ Dynamic[ performReplacements]], Spacer[ 5], Text[ translate[ "applyRepl"]]}], 
						translate[ "ttApplyRepl"]
				]};
		RRPDialogInput[ Notebook[ { 
							Cell[ BoxData[ ToBoxes[ InputField[ Dynamic[ expr], Boxes, FieldSize -> Medium]]], "Text", Deployed -> True], 
							Cell[ BoxData[ RowBox[ Join[ {ToBoxes[ Text[ translate["availConst"] <> ":"]], ToBoxes[ Spacer[ 5]]}, fixBut]]], "Text", ShowStringCharacters -> False, Deployed -> True],
							Cell[ TextData[ translate[ "proofSituation"]], "Subsubsection"],
							makeFormulaHighlightCells[ g, kb, Null, Null]
						}], 
						Hold[ {RRPBoxesToExpression[ expr], caseDistinction, performReplacements}], 
						"CloseOnEscape" -> False, "AdditionalMenuItems" -> menu, opts
		]
	]
	
formulaAbbrevDialog[ g_, kb_List, const_List, opts___?OptionQ] :=
	Module[ {sym = Null, expr = Null, equal = True,
				fixBut = Map[ ToBoxes[ PasteButton[ theoremaDisplay[ RNG$[ #]],
							With[ {fbox = ToBoxes[ First[ #], TheoremaForm], fc = First[ #]},
								DisplayForm[ InterpretationBox[ fbox, fc]]]]
							]&, const]},
		RRPDialogInput[ Notebook[ {
							Cell[ BoxData[ ToBoxes[ Row[ {InputField[ Dynamic[ sym], Boxes, FieldSize -> Small], Toggler[ Dynamic[ equal], {True -> Text[ ":="], False -> Text[ ":\[DoubleLeftRightArrow]"]}]}, Spacer[ 10]]]], "Text", ShowStringCharacters -> False, Deployed -> True],
							Cell[ BoxData[ ToBoxes[ InputField[ Dynamic[ expr], Boxes, FieldSize -> Medium]]], "Text", Deployed -> True], 
							Cell[ BoxData[ RowBox[ Join[ {ToBoxes[ Text[ translate["availConst"] <> ":"]], ToBoxes[ Spacer[ 5]]}, fixBut]]], "Text", ShowStringCharacters -> False, Deployed -> True],
							Cell[ TextData[ translate[ "proofSituation"]], "Subsubsection"],
							makeFormulaHighlightCells[ g, kb, Null, Null]
						}],
						Hold[ {RRPBoxesToExpression[ sym], RRPBoxesToExpression[ expr], equal}],
						"CloseOnEscape" -> False, opts
		]
	]

isSingleExpression[ expr_Hold] :=
	!MatchQ[ expr, Hold[ SEQ$[]|SEQ$[ _, __]|(VAR$|FIX$|META$)[ _SEQ0$|_SEQ1$, ___]]]
isSingleExpression[ expr_] :=
	isSingleExpression[ Hold[ expr]]
	
splitForallDialog[ g:(FML$[ _, Forall$TM[ rng_RNG$, __], __]), kb_List, opts___?OptionQ] :=
	Module[ {num = 1, simplify = True, menu},
		menu = {Tooltip[ Row[ {Checkbox[ Dynamic[ simplify]], Spacer[ 5], Text[ translate[ "simplifyForall"]]}], translate[ "ttSimplifyForall"]]};
		RRPDialogInput[ Notebook[ {
							Cell[ BoxData[ ToBoxes[ PopupMenu[ Dynamic[ num], MapIndexed[ (First[ #2] -> theoremaDisplay[ RNG$[ #1]])&, Most[ List@@rng]]]]], "Text", Deployed -> True],
							Cell[ TextData[ translate[ "proofSituation"]], "Subsubsection"],
							makeFormulaHighlightCells[ g, kb, Blank[], Null]
						}],
						Hold[ {num, simplify}],
						"CloseOnEscape" -> False, "AdditionalMenuItems" -> menu, opts
		]
	]


(* ::Subsection::Closed:: *)
(* Rewrite Rules *)


(* ::Subsubsection::Closed:: *)
(* Generating Rewrite Rules *)


$backwardGround = (True|False);

$ruleGenerationOptions = {
							Theorema`Language`HORewriting`Rules -> {"Backward", "Substitution", "ExplicitDefinition", "ImplicitDefinition", "LazyDefinition"}, (* no need to generate forward rules *)
							Theorema`Language`HORewriting`RightToLeft -> False,
							Theorema`Language`HORewriting`LazySymbols -> {},
							Theorema`Language`HORewriting`ImplicitCheckingFunction -> None,
							Theorema`Language`HORewriting`NormalizeArgumentLists -> "Lengthen",
							Theorema`Language`HORewriting`DecomposeEquivalences -> False
						 };

$extractTagsForRewriting = False;

trimFormulaForRewriting[ form:FML$[ _, And$TM[ __], ___]] :=
	(* We don't generate rewrite rules for conjunctions, but wait until they are split. *)
	{{form}}
	
If[ TrueQ[ $hoLoaded],
	trimFormulaForRewriting[ form_FML$] :=
	    Module[ {repl, kbForm, types, tags = {}, notIR = True, ambr = "All"},
	    	If[ TrueQ[ $extractTagsForRewriting],
	    		tags = getOptionalComponent[ form, "tags"];
	    		notIR = !MemberQ[ tags, "IR"];
	    		If[ MemberQ[ tags, "FO"], ambr = "Explicit"]
	 		];
	    	types = Replace[ Theorema`Language`HORewriting`Rules, $ruleGenerationOptions];
			repl = MapThread[ Rule,
	    					   {
	    						types,
	    						Theorema`Language`HORewriting`formulaToHORules @@ Join[ {form,
	    							Theorema`Language`HORewriting`ApplicationModuloBetaReduction -> ambr,
	    							Theorema`Language`HORewriting`BackwardWholeFormula -> notIR}, $ruleGenerationOptions]
	    					   }
	    			];
	    	kbForm = If[ notIR && (FreeQ[ formula@form, _IffDef$TM|EqualDef$TM[ _, _]] ||
	    							With[ {l = label@form}, StringLength[ l] > 6 && StringTake[ l, 6] === "Abbrv\[NumberSign]"]),
	    				{form},
	    				{}
					];
	    	Prepend[ repl, kbForm]
	    ];
	preprocessExpr = Theorema`Language`HORewriting`preprocessExpression;
	postprocessExpr = Theorema`Language`HORewriting`postprocessExpression;
	genGroundRule = Theorema`Language`HORewriting`generateGroundRule,
	
	(* else *)
	    
	trimFormulaForRewriting[ form_FML$] :=
		Module[ {kbForm, kbRules, goalRules, sRules, dRules, types},
			{kbForm, kbRules, goalRules, sRules, dRules} = Theorema`Language`FormulaManipulation`Private`trimFormulaForRewriting[ form];
			types = Replace[ Theorema`Language`HORewriting`Rules, $ruleGenerationOptions];
			If[ !MemberQ[ types, "Forward"], kbRules = {}];
			If[ !MemberQ[ types, "Backward"], goalRules = {}];
			If[ !MemberQ[ types, "Substitution"], sRules = {}];
			If[ !MemberQ[ types, "ExplicitDefinition"], dRules = {}];
			{kbRules, goalRules, sRules, dRules} =
				Replace[ {kbRules, goalRules, sRules, dRules}, {key_, rule:(lhs_ :> _)} :> {key, rule, FreeQ[ Hold[ lhs], _Blank|_BlankSequence|_BlankNullSequence]}, {2}];
			{kbForm, "Forward" -> kbRules, "Backward" -> goalRules, "Substitution" -> sRules, "ExplicitDefinition" -> dRules}
		];
	preprocessExpr[ x_] := x;
	postprocessExpr[ x_] := x;
	genGroundRule[ lhs_, rhs_, conds_, ___] :=
		With[ {out = Theorema`Language`FormulaManipulation`Private`mmaTransRule[ {lhs, rhs, conds, {}}, Null]},
			If[ out === $Failed,
				{},
			(*else*)
				{Null, out, True}
			]
		]
]


(* ::Subsubsection::Closed:: *)
(* Applying Rewrite Rules*)


(* 'RIReplace[ gRules, opts][ expr, repl]' rewrites 'expr' using the transformation rules in 'repl'.
	The following options can be given:
	- 'AcceptTrue': Boolean value indicating whether the result may be a formula whose truth value is 'True'.
	- 'CheckConditions': Boolean value indicating whether conditions should be checked
		(using 'followsImmediately' and 'gRules'), or not.
	- 'Except': List of *formulas* the result of rewriting must differ from.
	- 'ExceptHeads': List of symbols such that subexpressions of 'expr' whose head appears in the list are only
		rewritten as a whole, but not in parts.
	- 'Mode': 'Mode' can be any of 'First' (default), 'All', 'Repeated' and 'List'.
		With 'First', 'expr' is traversed in breadth-first manner, and for each subexpression the rules in 'repl'
		are tried in order. The first applicable rule is applied, and then the whole process stops. Hence, at most
		one rule will be applied in this mode.
		Mode 'All' imitates the behaviour of built-in 'ReplaceAll', i.e. all subexpressions of 'expr' are traversed in
		breadth-first manner, the first applicable rule is applied on each subexpression, and then the process continues
		with the next subexpression.
		Mode 'Repeated' imitates the behaviour of built-in 'ReplaceRepeated'.
		Mode 'List' applies each rule to each subexpression and returns a list of possible rewrites. The only difference
		to built-in 'ReplaceList' is that rules are applied only in one possible way, not in all possible ways.
	- 'ProveConditions': Boolean value specifying whether rewrite-rules constrained by (unchecked) conditions shall still
		be applied, causing these conditions to be proved in separate sub-proofs.
		Note that 'ProveConditions' and 'CheckConditions' are completely independent of each other; if both are 'False', this
		just means that only unconditional rules may be applied.
		
	The function returns a triple (or, with mode 'List', a list of triples) of the form
	'{expr, {refs, condrefs}, conds}'
	where
	- 'expr' is the new expression,
	- 'refs' are the formula references used for obtaining 'expr',
	- 'condrefs' are the formula references needed to justify rewrites (references of conditions; empty if 'CheckConditions' is 'False'),
	- 'conds' is a list of conditions that still have to be checked.
	
	Since we might want to pass 'RIReplace' to 'hoRewrite', and 'hoRewrite' only expects a binary function, we
	parameterize 'RIReplace' over the optional arguments and the rules for checking "followsImmediately".
	
	Attention!
	Specifying a non-empty list for 'Except' together with mode 'All' or 'Repeated' might lead to unexpected behaviour:
	Not only must the final expression be different from all 'Except'-formulas, but also all intermediate expressions!
*)
Options[ RIReplace] = {AcceptTrue -> True, Except -> {}, ExceptHeads -> {}, CheckConditions -> True, ProveConditions -> True, Mode -> First};
RIReplace[ _List, ___, Mode -> List, ___][ _, {}] := {}
RIReplace[ _List, ___][ expr_, {}] := {expr, {{}, {}}, {}}
RIReplace[ gRules_List, opts___?OptionQ][ orig_, repl_List] :=
	Module[ {i, expr = orig, tmpExpr = $Canceled, pos, new, uc, at, ex, eh, cc, pc, md, sub, tmp, out = {}, allRefs = {}, allCondRefs = {}, allConds = {}, conds},
		{at, ex, eh, cc, pc, md} = {AcceptTrue, Except, ExceptHeads, CheckConditions, ProveConditions, Mode} /. {opts} /. Options[ RIReplace];
		eh = Join[ {(VAR$|SEQ0$|SEQ1$)[ ___] -> VAR$, (FIX$|META$)[ ___] -> FIX$}, Map[ Rule[ Blank[ #], #]&, eh]];
		
		While[ tmpExpr === $Canceled || (md === Repeated && expr =!= tmpExpr),
			tmpExpr = expr;
			pos = Sort[ Position[ expr /. eh, Except[ VAR$], Heads -> True]];
			(* Built-in "Sort" automatically sorts the positions in breadth-first order. *)
			While[ pos =!= {},
				Catch[
					i = 0;
					Scan[
						Function[ p,
							++i;
							sub = myExtract[ expr, p];
							If[ !MatchQ[ sub, _VAR$|_SEQ0$|_SEQ1$|_RNG$|_SIMPRNG$|_STEPRNG$|_SETRNG$|_PREDRNG$|RNG$|SIMPRNG$|STEPRNG$|SETRNG$|PREDRNG$],
								Scan[
									Function[ r,
										{new, uc} = Reap[ Replace[ sub, r], {"ref", "cond"}];
										If[ uc =!= {{}, {}},
											new = myReplacePart[ expr, p -> new, Heads -> True];
											If[ (at || !TrueQ[ computeTruthValue[ new]]) && !memberOfKB[ ex, new],
												conds = DeleteCases[ uc[[2, 1]], True];
												tmp = If[ cc,
														followsImmediately[ conds, gRules, "RenameBound" -> False, "Strict" -> !pc],
														If[ pc || conds === {}, {{}, conds}, False]
													];
												If[ MatchQ[ tmp, {_List, _List}],
													If[ md === List,
														AppendTo[ out, {new, {uc[[1, 1]], DeleteDuplicates[ First[ tmp]]}, DeleteDuplicates[ Last[ tmp]]}],
														
														allRefs = DeleteDuplicates[ Join[ allRefs, uc[[1, 1]]]];
														allCondRefs = DeleteDuplicates[ Join[ allCondRefs, First[ tmp]]];
														allConds = DeleteDuplicates[ Join[ allConds, Last[ tmp]]];
														expr = new;
														If[ md === First,
															pos = {},
															With[ {patt = Append[ p, BlankNullSequence[]]},
																(* Remove all subpositions *)
																pos = DeleteCases[ Drop[ pos, i], patt];
															]
														];
														Throw[ Null]
													]
												]
											]
										]
									],
									repl
								]
							]
						],
						pos
					];
					pos = {}
				]
			]
		];
		If[ md === List,
			out,
			{expr, {allRefs, allCondRefs}, allConds}
		]
	]
	
(* "followsImmediately" is a method for finding out whether a certain formula follows 'immediately' from
	a given knowledge base / backward rules. It is mainly used in inference rules for rewriting when some
	rewrite rules have conditions; If the conditions follow 'immediately', the rewrite rule can be applied,
	otherwise not. *)
Options[ followsImmediately] = {"RenameBound" -> True, "Strict" -> True, "ConditionGenerator" -> Function[ {}], "Depth" -> 3};
followsImmediately[ f_, gRules_List, opts___?OptionQ] :=
	Module[ {rb, st, cg, dp},
		{rb, st, cg, dp} = {"RenameBound", "Strict", "ConditionGenerator", "Depth"} /. {opts} /. Options[ followsImmediately];
		followsImmediately[ f, gRules, rb, st, cg, dp]
	];
followsImmediately[ FML$[ _, f_, __], gRules_List, renameBound_, strict_, cg_, rest___] :=
	followsImmediately[ f, gRules, renameBound, strict, cg, rest]
followsImmediately[ (And$TM|List)[ args___], gRules_List, renameBound_, strict_, cg_, rest___] :=
	Module[ {refs = {}, conds = {}, tmp},
		Catch[
			Scan[(
				tmp = followsImmediately[ #, gRules, renameBound, strict, cg, rest];
				Switch[ tmp,
					{_List, _List},
					refs = Join[ refs, First[ tmp]];
					conds = Join[ conds, Last[ tmp]],
					_,
					Throw[ False]
				]
				)&,
				{args}
			];
			{refs, conds}
		]
	]
followsImmediately[ f_, gRules_List, renameBound_, strict_, cg_, dp_:3] :=
	Module[ {search, pp},
		pp = If[ TrueQ[ renameBound],
			 	preprocessExpr[ f],
			 	f
			];
		(* abort after 3 seconds *)
		search = TimeConstrained[ exhaustiveSearch[ pp, True, gRules, computeTruthValue, cg, dp], 3];
		Switch[ search,
			{True, _List, _List},
			If[ TrueQ[ renameBound],
				{search[[2]], postprocessExpr[ Last[ search]]},
				Rest[ search]
			],
			_,
			If[ TrueQ[ strict],
				False,
				{{}, {f}}
			]
		]
	]
	
exhaustiveSearch[ start_, target_, gRules_List, simp_, cg_, maxDepth_Integer] :=
	exhaustiveSearch[ target, gRules, simp, cg, {{start, {}}}, {}, maxDepth]
exhaustiveSearch[ target_, _List, _, _, {___, l:{target_, _List}, ___}, _List, _] := Append[l, {}]
exhaustiveSearch[ target_, gRules_List, simp_, cg_, l:{{first_, refs_List}, rest___}, visited_List, maxDepth_Integer] /; Length[ refs] <= maxDepth :=
	Module[ {new, newExpressions = {}, newNodes = {}, i, expri, conds},
		new = DeleteCases[ Map[ rewriteAllPositiveParts[ first, #, visited]&, gRules], {_, {}}];
		If[ new === {} && TrueQ[ target] && (conds = cg[ first]) =!= {} && alphaDisjointQ[ $forbiddenConditions, conds],
			{True, refs, conds},
			
			For[ i = 1, i <= Length[ new], i++,
				expri = Last[ new[[i]]];
				expri = DeleteDuplicates[ Join[ expri, Select[ Map[ simp, expri], !alphaMemberQ[ visited, #]&]]];
				newExpressions = Join[ newExpressions, expri];
				newNodes = Join[ newNodes, Map[ {#, Join[ refs, First[ new[[i]]]]}&, expri]]
			];
			exhaustiveSearch[ target, gRules, simp, cg, Join[ {rest}, newNodes], Join[ visited, newExpressions], maxDepth]
		]
	]
exhaustiveSearch[ _, _List, _, _, _List, _List, _] := $Failed
	
(* "rewriteAllPositiveParts" rewrites the positive parts of the expressions,
	i.e. those occuring inside an even number of Not$TM (also hidden ones, e.g. left-hand-side of Implies$TM).
	The whole expression is assumed to be positive, i.e. without an "invisible" Not$TM wrapped around.
*)
rewriteAllPositiveParts[ expr_, rule_, exceptions_List] :=
	rewriteAt[ expr, rule, exceptions, positivePositions[ expr /. {(VAR$|SEQ0$|SEQ1$)[ ___] -> VAR$, (FIX$|META$)[ ___] -> FIX$}, Except[ VAR$], {}, True]]
	
rewriteAt[ expr_, rule_, exceptions_List, pos_List] :=
	Module[ {newExpr, ref = $Failed, r, cond, out = {}, tmp},
		Scan[
			(
				tmp = myExtract[ expr, #];
				{newExpr, {r, cond}} = Reap[ Replace[ tmp, rule], {"ref", "cond"}];
				If[ newExpr =!= tmp && (cond === {} || cond === {{True}}),
					If[ r =!= {}, ref = First[ r]];
					tmp = myReplacePart[ expr, # -> newExpr];
					If[ !alphaMemberQ[ exceptions, tmp], AppendTo[ out, tmp]]
				]
			)&,
			pos
		];
		{If[ ref === $Failed, {}, ref], out}
	]

positivePositions[ expr:(Not$TM[ A_]), pattern_, {path___}, True] :=
	Module[ {out = positivePositions[ A, pattern, {path, 1}, False]},
		If[ MatchQ[ expr, pattern], PrependTo[ out, {path}]];
		out
	]
positivePositions[ expr:(Not$TM[ A_]), pattern_, {path___}, False] :=
	positivePositions[ A, pattern, {path, 1}, True]
positivePositions[ expr:((And$TM|Or$TM)[ args__]), pattern_, path_List, True] :=
	Module[ {out = Apply[ Join, MapIndexed[ positivePositions[ #1, pattern, Append[ path, First[ #2]], True]&, {args}]]},
		If[ MatchQ[ expr, pattern], PrependTo[ out, path]];
		out
	]
positivePositions[ expr:((And$TM|Or$TM)[ args__]), pattern_, {path___}, False] :=
	Apply[ Join, MapIndexed[ positivePositions[ #1, pattern, {path, First[ #2]}, False]&, {args}]]
positivePositions[ expr:(Implies$TM[ l_, r_]), pattern_, path_List, True] :=
	Module[ {out = Join[ positivePositions[ l, pattern, Append[ path, 1], False], positivePositions[ r, pattern, Append[ path, 2], True]]},
		If[ MatchQ[ expr, pattern], PrependTo[ out, path]];
		out
	]
positivePositions[ expr:(Implies$TM[ l_, r_]), pattern_, {path___}, False] :=
	Join[ positivePositions[ l, pattern, {path, 1}, True], positivePositions[ r, pattern, {path, 2}, False]]
positivePositions[ expr:(Forall$TM[ _, cond_, A_]), pattern_, path_List, True] :=
	Module[ {out = Join[ positivePositions[ cond, pattern, Append[ path, 2], False], positivePositions[ A, pattern, Append[ path, 3], True]]},
		If[ MatchQ[ expr, pattern], PrependTo[ out, path]];
		out
	]
positivePositions[ expr:(Forall$TM[ _, cond_, A_]), pattern_, {path___}, False] :=
	Join[ positivePositions[ cond, pattern, {path, 2}, True], positivePositions[ A, pattern, {path, 3}, False]]
positivePositions[ expr:(Exists$TM[ _, cond_, A_]), pattern_, path_List, True] :=
	Module[ {out = Join[ positivePositions[ cond, pattern, Append[ path, 2], True], positivePositions[ A, pattern, Append[ path, 3], True]]},
		If[ MatchQ[ expr, pattern], PrependTo[ out, path]];
		out
	]
positivePositions[ expr:(Exists$TM[ _, cond_, A_]), pattern_, {path___}, False] :=
	Join[ positivePositions[ cond, pattern, {path, 2}, False], positivePositions[ A, pattern, {path, 3}, False]]
positivePositions[ expr_, pattern_, path_List, pos_] :=
	If[ pos && MatchQ[ expr, pattern], {path}, {}]
	
(* "computeTruthValue" provides a very elementary mechanism to compute the truth values of formulae. *)
computeTruthValue[ Not$TM[ True]] := False
computeTruthValue[ Not$TM[ False]] := True
computeTruthValue[ And$TM[ True...]] := True
computeTruthValue[ And$TM[ ___, False, ___]] := False
computeTruthValue[ Or$TM[ False...]] := False
computeTruthValue[ Or$TM[ ___, True, ___]] := True
computeTruthValue[ Implies$TM[ a_, a_]] := True
computeTruthValue[ Implies$TM[ False, _]] := True
computeTruthValue[ Implies$TM[ _, True]] := True
computeTruthValue[ Implies$TM[ True, False]] := False
computeTruthValue[ (Iff$TM|IffDef$TM)[ a_, a_]] := True
computeTruthValue[ (Iff$TM|IffDef$TM)[ False, True]] := False
computeTruthValue[ (Iff$TM|IffDef$TM)[ True, False]] := False
computeTruthValue[ (Equal$TM|EqualDef$TM)[ a_, a_]] := True
computeTruthValue[ Forall$TM[ _, a_, a_]] := True
computeTruthValue[ Forall$TM[ _, False, _]] := True
computeTruthValue[ Forall$TM[ _, _, True]] := True
computeTruthValue[ Exists$TM[ _, False, _]] := False
computeTruthValue[ Exists$TM[ _, _, False]] := False
computeTruthValue[ f:((Not$TM|And$TM|Or$TM|Implies$TM|Iff$TM|IffDef$TM|Forall$TM|Exists$TM)[ args___])] :=
	Module[ {res},
		computeTruthValue[ res] /; (res = Map[ computeTruthValue, f]) =!= f
	]
computeTruthValue[ f_] := f


(* ::Subsection::Closed:: *)
(* Rewrite-Rules Data Structure *)


(* {ruleSet1 -> {...}, ruleSet2 -> {...}, ...} *)

addRulesToRW[ data_List, {}] := data
addRulesToRW[ data_List, rl:{{}, ___}] := addRulesToRW[ data, DeleteCases[ rl, {}]]
addRulesToRW[ data_List, {rl_List, rest___}] :=
	addRulesToRW[
		Map[ (First[ #] -> Join[ Replace[ First[ #], List@@Append[ data, First[ #] -> {}]], Last[ #]])&, rl],
		{rest}
	]
	
Options[ getRulesFromRW] = {"ground" -> (True|False), "keys" -> Blank[], "type" -> Blank[]};
getRulesFromRW[ data_List, opts___?OptionQ] :=
	getRuleClassesFromRW[ data, opts][[All, 2]]
	
getRuleClassesFromRW[ data_List, opts___?OptionQ] :=
	Module[ {g, k, t},
		{g, k, t} = {"ground", "keys", "type"} /. {opts} /. Options[ getRulesFromRW];
		Cases[ Flatten[ Cases[ data, (t -> l_) :> l], 1], {k, _, g, ___}]
	]
removeRulesFromRW[ data_List, opts___?OptionQ] :=
	Module[ {g, k, t},
		{g, k, t} = {"ground", "keys", "type"} /. {opts} /. Options[ getRulesFromRW];
		Replace[ data, (tp:t -> l_) :> (tp -> DeleteCases[ l, {k, _, g, ___}]), {1}]
	]
	



(* ::Subsection::Closed:: *)
(* General Purpose *)


(* The position specification of the entire expression, as returned by "Position", is {}.
	This, however, cannot be used with functions "Extract" and "ReplacePart", so we define it ourselves. *)
myExtract[ expr_, {}] := expr
myExtract[ expr_, pos_List] := Extract[ expr, pos]
myReplacePart[ _, (Rule|RuleDelayed)[ {}, new_], ___?OptionQ] := new
myReplacePart[ expr_, repl:(_Rule|_RuleDelayed), opts___?OptionQ] := ReplacePart[ expr, repl, opts]

sinkNegation[ Not$TM[ True], ___] := False
sinkNegation[ Not$TM[ False], ___] := True
sinkNegation[ Not$TM[ Not$TM[ a_]], ___] := sinkNegation[ a]
sinkNegation[ Not$TM[ And$TM[ a__]], ___] := sinkNegation[ Map[ Not$TM, Or$TM[ a]]]
sinkNegation[ Not$TM[ Or$TM[ a__]], ___] := sinkNegation[ Map[ Not$TM, And$TM[ a]]]
sinkNegation[ Not$TM[ Implies$TM[ a_, b_]], ___] := And$TM[ a, sinkNegation[ Not$TM[ b]]]
sinkNegation[ Not$TM[ Iff$TM[ a_, b_]], ___] := Or$TM[ And$TM[ a, sinkNegation[ Not$TM[ b]]], And$TM[ sinkNegation[ Not$TM[ a]], b]]
sinkNegation[ Not$TM[ Forall$TM[ r_RNG$, cond_, f_]], ___] := Exists$TM[ r, sinkNegation[ cond], sinkNegation[ Not$TM[ f]]]
sinkNegation[ Not$TM[ Exists$TM[ r_RNG$, cond_, f_]], ___] := Forall$TM[ r, sinkNegation[ cond], sinkNegation[ Not$TM[ f]]]
sinkNegation[ a:(_And$TM|_Or$TM|_Implies$TM|_Iff$TM|_IffDef$TM), rest___] := Map[ sinkNegation[ #, rest]&, a]
sinkNegation[ (h:Forall$TM|Exists$TM)[ r_RNG$, cond_, f_], rest___] := h[ r, sinkNegation[ cond, rest], sinkNegation[ f, rest]]
sinkNegation[ a_] := a
sinkNegation[ _, False] := $Failed

alphaMemberQ[ l_List, expr_] :=
	MemberQ[ l, _?(alphaEquivalent[ #, expr, {}]&)]
	
alphaDisjointQ[ {}, _List] := True
alphaDisjointQ[ a_List, b_List] :=
	Catch[
		Scan[ If[ alphaMemberQ[ a, #], Throw[ False]]&, b];
		True
	]

Options[ addNewKB] = {Mode -> Prepend};
addNewKB[ kb:{___FML$}, FML$[ _, True|Not$TM[ False], ___], opts___?OptionQ] := kb
addNewKB[ kb:{___FML$}, fml_FML$, opts___?OptionQ] :=
	Module[ {m},
		m = Mode /. {opts} /. Options[ addNewKB];
		If[ Catch[ Scan[ If[ alphaEquivalent[ formula@fml, formula@#], Throw[ True]]&, kb]; False],
			kb,
		(* else *)
			If[ TrueQ[ $autoGenerateRules],
				With[ {trimmed = trimFormulaForRewriting[ fml]},
					AppendTo[ $rewriteRules, Rest[ trimmed]];
					If[ m === Append,
						Join[ kb, First[ trimmed]],
					(*else*)
						Join[ First[ trimmed], kb]
					]
				],
			(*else*)
				m[ kb, fml]
			]
		]
	]
addNewKB[ args___] := unexpected[ addNewKB, {args}]

joinNewKB[ kb:{___FML$}, {}, ___?OptionQ] := kb
joinNewKB[ kb:{___FML$}, new:{__FML$}, opts___?OptionQ] :=
	Switch[ Mode /. {opts} /. Options[ addNewKB],
		Prepend,
		Fold[ addNewKB[ #1, #2, opts]&, kb, Reverse[ new]],
		_,
		Fold[ addNewKB[ #1, #2, opts]&, kb, new]
	]
joinNewKB[ args___] := unexpected[ joinNewKB, {args}]

memberOfKB[ kb_List, fml_FML$] :=
	memberOfKB[ kb, formula@fml]
memberOfKB[ kb_List, f:Except[ _FML$]] :=
	Catch[ Scan[ If[ alphaEquivalent[ f, formula@#], Throw[ True]]&, kb]; False]
	
RRPToBeProved[ data___?OptionQ] := Theorema`Provers`Common`Private`checkProofSuccess[ RRPMakePRFSIT[ data]]
RRPToBeProved[ args___] := unexpected[ RRPToBeProved, {args}]

Options[ RRPMakePRFSIT] = {goal :> makeRRPFML[], kb -> {}, id :> ToString[ Unique[ "PRFSIT$"]], rules -> Hold[],
	ruleActivity -> {}, rulePriority -> {}, strategy -> Identity, "rewriting" -> {}, "newAssms" -> {}};
RRPMakePRFSIT[ data___?OptionQ] :=
	Module[{g, k, i, r, a, p, s, rw, pr},
		{g, k, i, r, a, p, s, rw, pr} =
			{goal, kb, id, rules, ruleActivity, rulePriority, strategy, "rewriting", "proposedRules"} /. {data} /. Options[ RRPMakePRFSIT];
		If[ TrueQ[ $autoGenerateRules],
			Assert[ MatchQ[ $rewriteRules, _List]];
			rw = addRulesToRW[ rw, $rewriteRules];
		];
		If[ !ListQ[ pr],
			pr = If[ ListQ[ $proposedRules], $proposedRules, {}]
		];
		PRFSIT$[ g, k, i, rules -> r, ruleActivity -> a, rulePriority -> p, strategy -> s,
			"proposedRules" -> pr, "rewriting" -> rw,
			Sequence @@ DeleteDuplicates[ Cases[ {data}, HoldPattern[ (Rule|RuleDelayed)[ Except[ "rewriting"|"proposedRules", _String], _]]], First[ #1] === First[ #2]&]
		]
	]
RRPMakePRFSIT[ args___] := unexpected[ RRPMakePRFSIT, {args}]

simplifyFormulaByComputation[ FML$[ key_, f_, rest___]] :=
	With[ {out = sequenceFlatten[ simplifyFormulaByComputation[ f]]},
		If[ isIndividual[ out],
			FML$[ key, out, rest],
		(*else*)
			$Failed
		]
	]
(* TODO: Simply replacing all occurrences of certain expressions is dangerous:
	If they appear as arguments of functions with Hold-attribute, no replacements should be performed.
	This, in particular, applies to the Abbrev-binder, because it pre-computes positions of its bound variables. *)
simplifyFormulaByComputation[ f_] :=
	computeInProof[ f] //.
		{Equal$TM[ a_, b_] /; buiActProve[ "Equal"] && alphaEquivalent[ a, b] :> True,
		Unequal$TM[ a_, b_] /; buiActProve[ "Equal"] && alphaEquivalent[ a, b] :> False,
		Unequal$TM[ a_, b_] :> Not$TM[ Equal$TM[ a, b]],
		NotElement$TM[ a_, b_] :> Not$TM[ Element$TM[ a, b]]}
	
(* "extractFormulaFromRule" takes a rule and tries to extract the formula the rule originated from by
	looking at subexpressions of the form "Sow[ fml_, "ref"]". If it succeeds, it returns the formula,
	otherwise it returns $Failed. *)
extractFormulaFromRule[ {_, r:(_Rule|_RuleDelayed), ___}] := extractFormulaFromRule[ r]
extractFormulaFromRule[ (Rule|RuleDelayed)[ _, r_]] :=
	Module[ {list},
		list = Cases[ Hold[ r], HoldPattern[ Sow[ fml:FML$[ _, _, __], "ref"]] :> fml, Infinity, 1];
		If[ Length[ list] === 1,
			First[ list],
		(*else*)
			$Failed
		]
	]
	

(* The following functions are used in goal rewriting to obtain only the "simplest" formulae *)
subFormula[ a_, a_] := True
subFormula[ a_And$TM, b_And$TM] := Complement[ a, b] === And$TM[]
subFormula[ a_And$TM, b_] := subFormula[ a, And$TM[ b]]
subFormula[ a_, b_And$TM] := subFormula[ And$TM[ a], b]
subFormula[ a_, b_] := subFormula[ And$TM[ a], And$TM[ b]]

simplestFormulae[ {}] := {}
simplestFormulae[ {a:{_FML$, _}, b:({_FML$, _}...)}] := simplestFormulae[ {a}, {b}]
simplestFormulae[ l_List, {a_, b___}] :=
	simplestFormulae[ updateSimplest[ l, a, {}, True], {b}]
simplestFormulae[ l_List, {}] := l

updateSimplest[ {s:{FML$[ _, f_, __], _}, r___}, t:{FML$[ _, a_, __], _}, new:{n___}, add_] :=
	If[ TrueQ[ subFormula[ a, f]],
		updateSimplest[ {r}, t, new, add],	(* s must be removed *)
		If[ TrueQ[ subFormula[ f, a]],
			updateSimplest[ {r}, t, {n, s}, False],	(* s must be kept, t must not be added *)
			updateSimplest[ {r}, t, {n, s}, add]	(* s must be kept *)
		]
	]
updateSimplest[ {}, _, l_List, False] := l
updateSimplest[ {}, a_, {n___}, True] := {n, a}

Options[ makeRRPFML] = {key :> Theorema`Language`FormulaManipulation`Private`defKey[],
					formula -> True,
					label :> Theorema`Language`FormulaManipulation`Private`defLabel[],
					"simplification" -> simplifyFormulaByComputation,
					"postprocessing" -> sequenceFlatten,
					"wfCheck" -> ((sequenceType[ #] === {1, True})&),
					"marker" -> ""};
makeRRPFML[ data___?OptionQ] :=
	Module[{k, f, l, s, m, fs, out, check, post},
		{k, f, l, s, m, check, post} = {key, formula, label, "simplification", "marker", "wfCheck", "postprocessing"} /. {data} /. Options[ makeRRPFML];
		fs = s[ f];
		If[ Head[ fs] === s, fs = f];
		fs = post[ fs];
		If[ TrueQ[ check[ fs]],
			If[ m =!= "" && !StringMatchQ[ l, m~~"\[NumberSign]"|"\[SpaceIndicator]"~~__],
				l = With[ {sep = If[ StringMatchQ[ l, NumberString], "\[NumberSign]", "\[SpaceIndicator]"]}, m <> sep <> l]
			];
			(* We cannot use 'makeFML', because we don't want 'standardFormQuantifier' to be called on 'fs'. *)
			out = Theorema`Language`FormulaManipulation`Private`makeTmaFml[ k, fs, l, f];
			AppendTo[ $generated, out];
			out,
		(*else*)
			Throw[ $Failed]
		]
	]
makeRRPFML[ args___] := unexpected[ makeRRPFML, {args}]

activatedQ[ x_] :=
	(!ListQ[ $knowledgeActivated] || MemberQ[ $knowledgeActivated, x])
goalActivated[] := ($goalActivated =!= False)

SetAttributes[ getInferenceRuleParams, HoldRest]
getInferenceRuleParams[ All, {}, dialog_] :=
	Which[
		MatchQ[ Theorema`Provers`$inferenceRuleParams, {__}],
		Theorema`Provers`$inferenceRuleParams,

		Theorema`Provers`$allowInteraction =!= False,
		dialog,

		True,
		$Failed
	]
getInferenceRuleParams[ All, {{def_, patt_}, r0___}, dialog_] :=
	Which[
		MatchQ[ Theorema`Provers`$inferenceRuleParams, {__}],
		With[ {r = {r0}},
		With[ {lp = Length[ Theorema`Provers`$inferenceRuleParams], ld = Length[ r] + 1, patts = ("_"|Last[ #])& /@ r, defs = r[[All, 1]]},
		With[ {p1 = Cases[ Position[ Theorema`Provers`$inferenceRuleParams, "_"|patt, {1}, Heads -> False], {k_} /; k + ld > lp :> k]},
			Catch[
				Scan[
					Function[ {i},
						If[ MatchQ[ Drop[ Theorema`Provers`$inferenceRuleParams, i], Take[ patts, lp - i]],
							Throw[
								Join[
									Take[ Theorema`Provers`$inferenceRuleParams, i - 1], 
									MapThread[
										If[ #1 === "_", #2, #1]&,
										{Drop[ Theorema`Provers`$inferenceRuleParams, i - 1], Prepend[ Take[ defs, lp - i], def]}
									],
									Drop[ defs, lp - i]
								]
							]
						]
					],
					p1
				];
				Join[ Theorema`Provers`$inferenceRuleParams, {def}, defs]
			]
		]]],
		
		Theorema`Provers`$allowInteraction =!= False,
		dialog,

		True,
		$Failed
	]
getInferenceRuleParams[ i_Integer, def_List, dialog_] :=
	Which[
		ListQ[ Theorema`Provers`$inferenceRuleParams],
		With[ {ld = Length[ def],
				j = Replace[ Position[ Theorema`Provers`$inferenceRuleParams, "_", {1}, 1, Heads -> False], {{{k_}} :> k - 1, {} -> Length[ Theorema`Provers`$inferenceRuleParams]}]},
			Which[
				j >= i,
				Take[ Theorema`Provers`$inferenceRuleParams, i],
				
				j + ld >= i,
				Join[
					Take[ Theorema`Provers`$inferenceRuleParams, i - ld],
					ReplacePart[
						def,
						Cases[
							Position[ Theorema`Provers`$inferenceRuleParams, Except[ "_"], {1}, Heads -> False],
							{k_} /; k + ld > i :> ((k + ld - i) -> Theorema`Provers`$inferenceRuleParams[[k]])
						]
					]
				],
				
				Theorema`Provers`$allowInteraction =!= False,
				dialog,
		
				True,
				$Failed
			]
		],

		Theorema`Provers`$allowInteraction =!= False,
		dialog,

		True,
		$Failed
	]
getInferenceRuleParams[ def_List, dialog_] :=
	Which[
		ListQ[ Theorema`Provers`$inferenceRuleParams],
		Which[
			MatchQ[ Theorema`Provers`$inferenceRuleParams, {Except[ "_"], ___}],
			First[ Theorema`Provers`$inferenceRuleParams],
			
			MatchQ[ def, {__}],
			First[ def],
			
			Theorema`Provers`$allowInteraction =!= False,
			dialog,
			
			True,
			$Failed
		],

		Theorema`Provers`$allowInteraction =!= False,
		dialog,
		
		True,
		$Failed
	]


(* ::Section::Closed:: *)
(* Propositional Logic *)


(* The following rules are almost exactly the same as the ones in "BasicTheoremaLanguage", with one important
	difference: Functions "addNewKB" and "joinNewKB" are used rather than "appendKB" and "joinKB".
	Note that the names of the inference rules are in most cases the ones of the rules in "BasicTheoremaLanguage",
	because the proof texts are identical anyway.
*)

inferenceRule[ RRPgoalInKB] =
PRFSIT$[ goal:FML$[ _, g_, __], {___, k:FML$[ _, a_, __], ___}, ___] /; alphaEquivalent[ g, a] :> performProofStep[
	makeTERMINALNODE[ makePRFINFO[ name -> goalInKB, used -> {goal, k}], proved]
]

inferenceRule[ RRPnotEquGoal] = 
PRFSIT$[ g:FML$[ _, Not$TM[ a:(Equal$TM[ _, __]|Iff$TM[ _, __])], __], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {opp},
		opp = makeRRPFML[ formula -> a, "marker" -> "A"];
		makeANDNODE[ makePRFINFO[ name -> notGoal, used -> g],
			RRPToBeProved[ goal -> makeRRPFML[ formula -> False, label -> "\[UpTee]"], kb -> addNewKB[ k, opp, Mode -> Prepend], "newAssms" -> {key@opp}, rest]
		]
	]
]

(*
	The pattern of the following rule must be like that, because even if the KB contains several conjunctions
	and 'ReplaceList' is used for application, there must only be 1 match.
*)
inferenceRule[ RRPandKB] = 
PRFSIT$[ g_, k_List?(MemberQ[ #, FML$[ _?activatedQ, And$TM[ __], __]]&), _, rest___?OptionQ] :> performProofStep[
	Block[ {$applyAllPossibleInferences = False},
		simplifyPSBySplitting[ g, k, And$TM[ __], splitConjunction, RRPandKB, rest]
	]
]

splitConjunction[ FML$[ _, And$TM[ f__], l_, ___]] :=
	MapIndexed[ makeRRPFML[ formula -> #1, label -> l <> "." <> ToString[ #2[[1]]], "marker" -> "A"]&, {f}]

inferenceRule[ RRPandGoal] = 
PRFSIT$[ g:FML$[ _, a:And$TM[ c__], lab_, ___] /; FreeQ[ a, _META$], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {conjG, conjA, knew = k, newAssms = {}},
		conjG = MapIndexed[ makeRRPFML[ formula -> #1, label -> lab <> "." <> ToString[ #2[[1]]], "marker" -> "G"]&, {c}];
		conjA = MapIndexed[ makeRRPFML[ formula -> #1, label -> lab <> "." <> ToString[ #2[[1]]], "marker" -> "A"]&, {c}];
		makeANDNODE[ makePRFINFO[ name -> RRPandGoal, used -> g],
			MapThread[ With[ {tmp = RRPToBeProved[ goal -> #1, kb -> knew, "newAssms" -> newAssms, rest]}, newAssms = {key@#2}; knew = addNewKB[ knew, #2]; tmp]&, {conjG, conjA}]
		]
	]
]

inferenceRule[ RRPorGoal] = 
PRFSIT$[ g:FML$[ _, o:Or$TM[ a__, b_], lab_, ___] /; FreeQ[ o, _META$], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {negAssum, newG},
		negAssum = MapIndexed[ makeRRPFML[ formula -> sinkNegation[ Not$TM[#1]], label -> lab <> "." <> ToString[ #2[[1]]], "marker" -> "A"]&, {a}];
		newG = makeRRPFML[ formula -> b, "marker" -> "G"];
		makeANDNODE[ makePRFINFO[ name -> orGoal, used -> g],
			RRPToBeProved[ goal -> newG, kb -> joinNewKB[ k, negAssum], "newAssms" -> key /@ negAssum, rest]
		]
	]
]

inferenceRule[ RRPorKB] = 
PRFSIT$[ g_, {pre___, k:FML$[ _?activatedQ, Or$TM[ c__], lab_, ___], post___}, _, rest___?OptionQ] :> performProofStep[
	Module[ {caseAssum, negAssum, knew = {pre, post}, tmp},
		caseAssum = MapIndexed[ makeRRPFML[ formula -> #1, label -> lab <> "." <> ToString[ #2[[1]]], "marker" -> "A"]&, {c}];
		negAssum = MapIndexed[ makeRRPFML[ formula -> sinkNegation[ Not$TM[ #1]], label -> lab <> "." <> ToString[ #2[[1]]] <> ".\[Not]", "marker" -> "A"]&, {c}];
		makeANDNODE[ makePRFINFO[ name -> RRPorKB, used -> k, generated -> caseAssum],
			MapThread[ (
							Block[ {$rewriteRules = $rewriteRules},
								tmp = RRPToBeProved[ goal -> g, kb -> addNewKB[ knew, #1], "newAssms" -> {key@#1}, rest]
							];
							knew = addNewKB[ knew, #2];
							tmp
						)&,
						{caseAssum, negAssum}
			]
		]
	]
]

inferenceRule[ RRPimplGoalDirect] = 
PRFSIT$[ g:FML$[ _, Implies$TM[ P_, Q_], __], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {left, right},
		left = makeRRPFML[ formula -> P, "marker" -> "A"];
		right = makeRRPFML[ formula -> Q, "marker" -> "G"];
		makeANDNODE[ makePRFINFO[ name -> implGoalDirect, used -> g],
			RRPToBeProved[ goal -> right, kb -> addNewKB[ k, left], "newAssms" -> {key@left}, rest]
		]
	]
]

(*
	The pattern of the following rule must be like that, because even if the KB contains several equivalences
	and 'ReplaceList' is used for application, there must only be 1 match.
*)
inferenceRule[ RRPequivKB] = 
PRFSIT$[ g_FML$, k_List?(MemberQ[ #, FML$[ _?activatedQ, Iff$TM[ _, _], __]]&), _, rest___?OptionQ] :> performProofStep[
	simplifyPSBySplitting[ g, k, Iff$TM[ _, _], splitEquivalence, RRPequivKB, rest]
]

splitEquivalence[ FML$[ _, Iff$TM[ lhs_, rhs_], l_, ___]] :=
	{makeRRPFML[ formula -> Implies$TM[ lhs, rhs], label -> l <> ".\[DoubleRightArrow]", "marker" -> "A"],
	 makeRRPFML[ formula -> Implies$TM[ rhs, lhs], label -> l <> ".\[DoubleLeftArrow]", "marker" -> "A"]}
	 
inferenceRule[ RRPmodusPonensTollens] =
this:PRFSIT$[ g_, k_List?(MemberQ[ #, FML$[ _, Implies$TM[ _, _], __]]&), _, rest___?OptionQ] :> performProofStep[
	Module[ {rw = getOptionalComponent[ this, "rewriting"], f, pos, delPos = {}, impl, assm, gRules,
				condRefs, conds, allCondRefs = {}, allConds = {}, newForm, tmp, action = "", usd = {}, gen = {}, ok},
		gRules = getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround];
		(* Since we really check *all* conditions in calls to 'followsImmediately' in this functions,
			i.e. do not take into account definedness/domain conditions, we do not need to set '$forbiddenConditions' either. *)
		pos = Position[ k, FML$[ _?activatedQ, Implies$TM[ _, _], __], {1}, Heads -> False];
		
		Catch[
			Scan[ Function[ p,
					ok = Catch[	(* catches all '$Failed' thrown by 'makeRRPFML' *)
						f = Extract[ k, p];
						impl = simplifiedImplies[ formula@f];
						assm = First[ impl];
						condRefs = {};
						conds = {};
						If[ action =!= "modus tollens",
							Switch[ assm,
								And$TM[ __],
								newForm = And$TM[];
								condRefs = {};
								Scan[ Function[ conj,
										tmp = followsImmediately[ conj, gRules];
										Switch[ tmp,
											{_List, _List},
											condRefs = Join[ condRefs, First[ tmp]];
											conds = Join[ conds, Last[ tmp]],
											_,
											AppendTo[ newForm, conj]
										]
									],
									assm
								];
								newForm = Which[ newForm === And$TM[],
												action = "modus ponens";
												makeRRPFML[ formula -> Last[ impl], "marker" -> "A"],
												
												condRefs =!= {} || conds =!= {},
												action = "modus ponens";
												makeRRPFML[ formula -> Implies$TM[ newForm, Last[ impl]], "marker" -> "A"],
												
												True,
												$Failed
											],
								
								_,
								tmp = followsImmediately[ assm, gRules];
								newForm = Switch[ tmp,
												{_List, _List},
												{condRefs, conds} = tmp;
												action = "modus ponens";
												makeRRPFML[ formula -> Last[ impl], "marker" -> "A"],
												_,
												$Failed
											]
							]
						];
						If[ newForm === $Failed && action =!= "modus ponens",
							assm = Not$TM[ Last[ impl]];
							tmp = followsImmediately[ assm, gRules];
							newForm = Switch[ tmp,
											{_List, _List},
											{condRefs, conds} = tmp;
											action = "modus tollens";
											makeRRPFML[ formula -> sinkNegation[ Not$TM[ First[ impl]]], "marker" -> "A"],
											_,
											$Failed
										]
						];
						True
					];
					If[ TrueQ[ ok] && newForm =!= $Failed,
						AppendTo[ delPos, p];
						allCondRefs = Join[ allCondRefs, condRefs];
						allConds = Join[ allConds, conds];
						AppendTo[ usd, f];
						AppendTo[ gen, newForm];
						(* The rewrite-rules which originated from the old formula 'f' are superfluous, because the ones that
							originate from the new formula 'newForm' have fewer conditions. *)
						rw = removeRulesFromRW[ rw, "keys" -> key@f];
						If[ !TrueQ[ $applyAllPossibleInferences], Throw[ Null]];
					]
				],
				pos
			]
		];
		If[ usd === {},
			$Failed,
		(*else*)
			Switch[ allConds,
				{},
				makeANDNODE[
					makePRFINFO[
						name -> RRPmodusPonensTollens,
						used -> usd,
						generated -> {gen, {}},
						"usedForms" -> DeleteDuplicates[ allCondRefs],
						"action" -> action
					], 
					RRPToBeProved[ goal -> g, kb -> joinNewKB[ Delete[ k, delPos], gen, Mode -> Prepend], "rewriting" -> rw, "newAssms" -> key /@ gen, rest]
				],
				
				{__},
				allConds = makeRRPFML[ formula -> And$TM@@allConds, "marker" -> "C"];
				makeANDNODE[
					makePRFINFO[
						name -> RRPmodusPonensTollens,
						used -> usd,
						generated -> {gen, {allConds, g}},
						"usedForms" -> DeleteDuplicates[ allCondRefs],
						"action" -> action
					],
					{RRPToBeProved[ goal -> allConds, kb -> Delete[ k, delPos], rest],
					RRPToBeProved[ goal -> g, kb -> joinNewKB[ Delete[ k, delPos], {allConds, newForm}, Mode -> Prepend], "rewriting" -> rw, "newAssms" -> {key@allConds, key@newForm}, rest]}
				]
			]
		]
	]
]
	 
(*
	The pattern of the following rule must be like that, because even if the KB contains several equivalences
	and 'ReplaceList' is used for application, there must only be 1 match.
*)
inferenceRule[ RRPsimpImplication] =
PRFSIT$[ g_FML$, k_List?(MemberQ[ #, FML$[ _?activatedQ, Implies$TM[ _Or$TM, _], __]]&), _, rest___?OptionQ] :> performProofStep[
	simplifyPSBySplitting[ g, k, Implies$TM[ _Or$TM, _], splitImplication, RRPsimpImplication, rest]
]

splitImplication[ FML$[ _, Implies$TM[ Or$TM[ p___], r_], l_, ___]] :=
	MapIndexed[ makeRRPFML[ formula -> Implies$TM[ #1, r], label -> l <> "." <> ToString[ #2[[1]]], "marker" -> "A"]&, {p}]
	
simplifyPSBySplitting[ g_, k_List, pattern_, fun_, n_Symbol, rest___?OptionQ] :=
	Module[ {forms, new, pos = Position[ k, FML$[ _?activatedQ, pattern, __], {1}, Heads -> False]},
		forms = Extract[ k, pos];
		
		If[ TrueQ[ $applyAllPossibleInferences],
			MapThread[ (
					$rewriteRules = {};
					new = fun[ #1];
					makeANDNODE[ makePRFINFO[ name -> n, used -> {#1}, generated -> {new}],
						RRPToBeProved[ goal -> g, kb -> joinNewKB[ Delete[ k, First[ #2]], new, Mode -> Prepend], "newAssms" -> key /@ new, rest]
					]
				)&,
				{forms, pos}
			],
		(*else*)
			new = Join @@ Map[ fun, forms];
			makeANDNODE[ makePRFINFO[ name -> n, used -> forms, generated -> new],
				RRPToBeProved[ goal -> g, kb -> joinNewKB[ Delete[ k, pos], new, Mode -> Prepend], "newAssms" -> key /@ new, rest]
			]
		]
	]
	
inferenceRule[ RRPcontradiction] =
PRFSIT$[ g:FML$[ _, a:Except[ False], __], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {fml},
		fml = makeRRPFML[ formula -> sinkNegation[ Not$TM[ a]], "marker" -> "A"];
		makeANDNODE[ makePRFINFO[ name -> contradiction, used -> g], 
			RRPToBeProved[ goal -> makeRRPFML[ formula -> False, label -> "\[UpTee]"], kb -> addNewKB[ k, fml], "newAssms" -> {key@fml}, rest]
		]
	]
]


(* ::Section::Closed:: *)
(* Miscellaneous *)

inferenceRule[ RRPsimplifyNegation] =
PRFSIT$[ g:FML$[ _, f_, __], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {simpGoal = False, newGoal},
		If[ goalActivated[],
			newGoal = sinkNegation[ f, False];
			If[ newGoal === $Failed,
				newGoal = g,
			(*else*)
				simpGoal = True;
				newGoal = makeRRPFML[ formula -> newGoal, "marker" -> "G"]
			],
		(*else*)
			newGoal = g
		];
		With[ {newAssms = Map[
					Function[ {assm},
						If[ activatedQ[ key@assm],
							With[ {new = sinkNegation[ formula@assm, False]},
								If[ new === $Failed,
									$Failed,
								(*else*)
									makeRRPFML[ formula -> new, "marker" -> "A"]
								]
							],
						(*else*)
							$Failed
						]
					],
					k
				]},
		With[ {simpPos = Position[ newAssms, _FML$, {1}, Heads -> False]},
			If[ simpPos =!= {} || simpGoal,
				With[ {simpAssms = Extract[ newAssms, simpPos], usedAssms = Extract[ k, simpPos]},
					makeANDNODE[
						makePRFINFO[ name -> RRPsimplifyNegation, used -> {If[ simpGoal, {g}, {}], usedAssms}, generated -> {If[ simpGoal, {newGoal}, {}], simpAssms}],
			        	RRPToBeProved[ goal -> newGoal, kb -> joinNewKB[ Delete[ k, simpPos], simpAssms], "newAssms" -> key /@ simpAssms, rest]
					]
				],
			(*else*)
				$Failed
			]
		]]
	]
]

inferenceRule[ RIexpandSuch] =
this:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {pos, fml, out, cond, r, c, f, fix, such, mainBranch, newForm,
				except = getOptionalComponent[ this, "RIexpandSuch"],
				rw = getOptionalComponent[ this, "rewriting"],
				locC = getOptionalComponent[ this, "constants"],
				data = {If[ goalActivated[], g, Null], Replace[ k, FML$[ _?(!activatedQ[ #]&), __] -> Null, {1}]}
			},
		pos = Position[ data,
						e:((Such$TM|SuchUnique$TM)[ _RNG$, _, _]) /; (!alphaMemberQ[ except, e] && freeVariables[ e] === {}),
						{1, Infinity},
						If[ TrueQ[ $applyAllPossibleInferences], Infinity, 1]
				];
		out = Map[ Function[ p,
						fml = If[ First[ p] === 1, g, k[[p[[2]]]]];
						such = Extract[ data, p];
						cond = Exists$TM@@such;
						If[ Head[ such] === SuchUnique$TM,
							cond = makeExistsUnique[ cond]
						];
						r = makeConjunction[ rngToCondition[ First[ cond]], And$TM];
						{{r, c, f}, fix} = arbitraryButFixed[ Prepend[ List@@Rest[ cond], r], First[ cond], {g, k, locC}];
						locC = Prepend[ locC, fix];
						fix = rngConstants[ fix];
						If[ Length[ fix] === 1,
							fix = First[ fix],
							fix = Tuple$TM@@fix	
						];
						{r, c, f} = Map[ makeRRPFML[ formula -> #, "marker" -> "A"]&, {r, c, f}];
						cond = makeRRPFML[ formula -> cond, "marker" -> "C"];
						
						If[ First[ p] === 1,
							newForm = makeRRPFML[ formula -> myReplacePart[ formula@fml, Drop[ p, 2] -> fix], "marker" -> "G"];
							mainBranch = Block[ {$rewriteRules = {}}, RRPToBeProved[ goal -> newForm, kb -> joinNewKB[ k, {r, c, f}], "constants" -> locC, "newAssms" -> key /@ {r, c, f}, rest]],
							
							rw = removeRulesFromRW[ rw, "keys" -> key@fml];
							newForm = makeRRPFML[ formula -> myReplacePart[ formula@fml, Drop[ p, 3] -> fix], "marker" -> "A"];
							mainBranch = Block[ {$rewriteRules = {}},
											RRPToBeProved[ goal -> g, kb -> joinNewKB[ Delete[ k, p[[2]]], {newForm, r, c, f}], "rewriting" -> rw, "constants" -> locC, "newAssms" -> key /@ {newForm, r, c, f}, rest]
										]
						];
						makeANDNODE[ makePRFINFO[ name -> RIexpandSuch, used -> fml, generated -> {cond, newForm, r, c, f}, "data" -> {p, such, fix}],
									{RRPToBeProved[ goal -> cond, kb -> k, "RIexpandSuch" -> Prepend[ except, such], rest], mainBranch}
						]
					],
					pos
				];
		Switch[ out,
			{},
			$Failed,
			{_},
			First[ out],
			_,
			out	
		]
	]
]

makeExistsUnique[ Exists$TM[ rng_RNG$, cond_, form_]] :=
	Module[ {rv = rngVariables[ rng], newVars, uc},
		newVars = freshName[ rv, freeVariables[ {List @@ rng, cond, form}] /. {VAR$[ (SEQ0$|SEQ1$)[ s_Symbol]] :> s, VAR$[ s_Symbol] :> s}];
		uc = renameBound[ Forall$TM[ rng, cond, form], newVars, "checkTypes" -> False, "postprocessing" -> Identity];
		uc = ReplacePart[ uc, -1 -> simplifiedImplies[ Implies$TM[ Last[ uc], simplifiedAnd[ And$TM @@ MapThread[ Equal$TM, {newVars, rv}]]]]];
		Exists$TM[ rng, cond, simplifiedAnd[ And$TM[ form, uc]]]
	]

inferenceRule[ RRPexpandPiecewise] =
this:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {rw = getOptionalComponent[ this, "rewriting"], pos, gRules, data},
		gRules = getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround];
		
		data = {If[ goalActivated[], formula@g, Null], Map[ If[ activatedQ[ key[ #]], formula[ #], Null]&, k]} /.
								{Piecewise$TM[ Tuple$TM[]] -> Null,
								pw_Piecewise$TM :> Piecewise$TM[ freeVariables[ pw]]};

		If[ TrueQ[ $applyAllPossibleInferences],
			Map[ expandPW[ g, k, #, gRules, rest]&, Position[ data, Piecewise$TM[ {}]]],
		(*else*)
			pos = Position[ data, Piecewise$TM[ {}], Infinity, 1];
			If[ pos === {},
				$Failed,
			(*else*)
				expandPW[ g, k, First[ pos], gRules, rest]
			]
		]
	]
]

expandPW[ g_, k_List, {1, r___}, gRules_List, rest___] := expandPWGoal[ g, k, {r}, gRules, rest]
expandPW[ g_, k_List, {2, i_, r___}, gRules_List, rest___] := expandPWKB[ g, k, i, {r}, gRules, rest]
	
expandPWGoal[ g_, k_List, pos_List, gRules_List, rest___] :=
	Module[ {pwOrig, pwNew, refs, newG, newK = {}, neg = {}, fml = formula@g, simp},
		pwOrig = myExtract[ fml, pos];
		{pwNew, refs, simp} = simplifyPW[ pwOrig, gRules];
		Switch[ pwNew,
			Piecewise$TM[ Tuple$TM[]],
			newG = makeRRPFML[ formula -> myReplacePart[ fml, pos -> pwNew], "marker" -> "G"];
			makeANDNODE[ makePRFINFO[ name -> RRPexpandPiecewise, used -> {{g}, refs}, generated -> {{newG}, {}}, "orig" -> pwOrig, "new" -> pwNew, "simp" -> simp, "place" -> "goal"],
	        		RRPToBeProved[ goal -> newG, kb -> k, rest]
			],
			
			Piecewise$TM[ Tuple$TM[ __]],
			newG = Map[ With[ {expr = First[ #], cond = Last[ #]},
							If[ cond === True,
								AppendTo[ newK, neg],
								AppendTo[ newK, Append[ neg, makeRRPFML[ formula -> cond, "marker" -> "A"]]];
								AppendTo[ neg, makeRRPFML[ formula -> Not$TM[ cond], "marker" -> "A"]]
							];
							makeRRPFML[ formula -> myReplacePart[ fml, pos -> expr], "marker" -> "G"]
						]&,
						List@@First[ pwNew]
					];
			If[ pwNew[[1, -1, 2]] =!= True,
				AppendTo[ newG, makeRRPFML[ formula -> myReplacePart[ fml, pos -> Piecewise$TM[ Tuple$TM[]]], "marker" -> "G"]];
				AppendTo[ newK, neg]
			];
			makeANDNODE[ makePRFINFO[ name -> RRPexpandPiecewise, used -> {{g}, refs}, generated -> {newG, newK}, "orig" -> pwOrig, "new" -> pwNew, "simp" -> simp, "place" -> "goal"],
	        		MapThread[ Block[ {$rewriteRules = {}},
	        						(* It is probably not the most efficient way to generate rewrite rules again and again for
	        							all the formulas that appear in multiple knowledge bases. However, those formulas are negated, i.e.
	        							have a 'Not$TM' as their head, so the only rewrite rules are ground backward rules. *)
	        						RRPToBeProved[ goal -> #1, kb -> joinNewKB[ k, #2], "newAssms" -> key /@ #2, rest]
		        				]&,
	        					{newG, newK}
	        		]
			],
			
			_,
			$Failed
		]
	]
	
expandPWKB[ g_, k_List, index_, pos_List, gRules_List, rest___] :=
	Module[ {pwOrig, pwNew, refs, newF, newK = {}, neg = {}, fml = formula@k[[index]], simp, thinnedKB = Delete[ k, index]},
		pwOrig = myExtract[ fml, pos];
		{pwNew, refs, simp} = simplifyPW[ pwOrig, gRules];
		Switch[ pwNew,
			Piecewise$TM[ Tuple$TM[]],
			newF = makeRRPFML[ formula -> myReplacePart[ fml, pos -> pwNew], "marker" -> "A"];
			makeANDNODE[ makePRFINFO[ name -> RRPexpandPiecewise, used -> {{k[[index]]}, refs}, generated -> {{newF}, {}}, "orig" -> pwOrig, "new" -> pwNew, "simp" -> simp, "place" -> "kb"],
	        		Block[ {$autoGenerateRules = False},
	        			(* Don't generate rewrite rules for a formula containing Piecewise[{}]. *)
	        			RRPToBeProved[ goal -> g, kb -> addNewKB[ thinnedKB, newF, Mode -> Prepend], "newAssms" -> {key@newF}, rest]
	        		]
			],
			
			Piecewise$TM[ Tuple$TM[ __]],
			newF = Map[ With[ {expr = First[ #], cond = Last[ #]},
							If[ cond === True,
								AppendTo[ newK, neg],
								AppendTo[ newK, Append[ neg, makeRRPFML[ formula -> cond, "marker" -> "A"]]];
								AppendTo[ neg, makeRRPFML[ formula -> Not$TM[ cond], "marker" -> "A"]]
							];
							makeRRPFML[ formula -> myReplacePart[ fml, pos -> expr], "marker" -> "A"]
						]&,
						List@@First[ pwNew]
					];
			If[ pwNew[[1, -1, 2]] =!= True,
				AppendTo[ newF, makeRRPFML[ formula -> myReplacePart[ fml, pos -> Piecewise$TM[ Tuple$TM[]]], "marker" -> "A"]];
				AppendTo[ newK, neg]
			];
			makeANDNODE[ makePRFINFO[ name -> RRPexpandPiecewise, used -> {{k[[index]]}, refs}, generated -> {newF, newK}, "orig" -> pwOrig, "new" -> pwNew, "simp" -> simp, "place" -> "kb"],
	        		MapThread[ Block[ {$rewriteRules = {}},
	        						(* It is probably not the most efficient way to generate rewrite rules again and again for
	        							all the formulas that appear in multiple knowledge bases. However, those formulas are negated, i.e.
	        							have a 'Not$TM' as their head, so the only rewrite rules are ground backward rules. *)
	        						RRPToBeProved[ goal -> g, kb -> addNewKB[ joinNewKB[ thinnedKB, #2], #1], "newAssms" -> Append[ key /@ #2, key@#1], rest]
		        				]&,
	        					{newF, newK}
	        		]
			],
			
			_,
			$Failed
		]
	]
	
simplifyPW[ Piecewise$TM[ t_Tuple$TM], gRules_List] :=
	Module[ {condRefs = {}, rules = gRules, cases, tc = False, c, tmp, fc, simp = False, replace = True},
		cases = DeleteCases[
					Map[ Function[ case, If[ tc,
						simp = True;
						Null,
						
						c = preprocessExpr[ Flatten[ And$TM[ Last[ case]]]];
						fc = Catch[ Scan[
										(
											tmp = followsImmediately[ Not$TM[ #], rules, "RenameBound" -> False];
											If[ MatchQ[ tmp, {_List, _List}],
												simp = True;
												condRefs = Join[ condRefs, First[ tmp]];
												rules = Join[ rules, genGroundRule[ Not$TM[ #], True, {}, "Preprocess" -> False][[2]]];
												Throw[ True]
											]
										)&,
										c
									];
									Throw[ False]
								];
						If[ fc,
							Null,
							c = Select[ c, (
												tmp = followsImmediately[ #, rules, "RenameBound" -> False];
												Switch[ tmp,
													{_List, _List},
													simp = True;
													condRefs = Join[ condRefs, First[ tmp]];
													rules = Join[ rules, genGroundRule[ #, True, {}, "Preprocess" -> False][[2]]];
													False,
													_,
													True
												]
											)&
								];
							If[ Length[ c] === 0,
								tc = True;
								Tuple$TM[ First[ case], True],
								
								If[ Length[ c] === 1, c = First[ c]];
								rules = Join[ rules, genGroundRule[ Not$TM[ c], True, {}, "Preprocess" -> False][[2]]];
								Tuple$TM[ First[ case], postprocessExpr[ c]]
							]
						]
					]],
					t
				],
				Null
			];
			
		If[ cases =!= Tuple$TM[],
			While[ replace,
				replace = False;
				cases = Replace[ cases, Tuple$TM[ pre___, Tuple$TM[ e1_, c1_], mid___, Tuple$TM[ e2_, c2_], post___] /; alphaEquivalent[ e1, e2] :>
										(
											replace = True;
											simp = True;
											Tuple$TM[ pre, Tuple$TM[ e1, simplifiedOr[ Or$TM[ c1, simplifiedAnd[
													Append[ Map[ sinkNegation[ Not$TM[ Last[#]]]&, And$TM[ mid]], c2]
												]]]],
												mid, post]
										)
						]
			];
			(* The previous simplification might have replaced a last 'e <= True'-case by something else.
				Therefore, we make sure that the last case condition becomes 'True' again. *)
			If[ tc && cases[[-1, 2]] =!= True, cases[[-1, 2]] = True]
		];
  
		{Piecewise$TM[ cases], condRefs, simp}
	]


(* ::Section::Closed:: *)
(* Quantifiers *)


inferenceRule[ RRPforallGoal] = 
ps:PRFSIT$[ g:FML$[ _, u:Forall$TM[ rng_, cond_, A_], __], k_List, _, rest___?OptionQ] :>
	Module[ {rc, r, c, f, fix, newConds, newGoal, locC},
		performProofStep[
			locC = getOptionalComponent[ ps, "constants"];
			{{r, c, f}, fix} = arbitraryButFixed[ {rc, cond, A}, rng, {g, k, locC}];
			newGoal = makeRRPFML[ formula -> f, "marker" -> "G"];
			newConds = Map[ makeRRPFML[ formula -> #, "marker" -> "A"]&, DeleteCases[ Append[ r, c], True]];
			makeANDNODE[ makePRFINFO[ name -> forallGoal, used -> g, "abf" -> rngConstants[ fix]], 
				RRPToBeProved[ goal -> newGoal, kb -> joinNewKB[ k, newConds, Mode -> Prepend], "constants" -> Prepend[ locC, fix], "newAssms" -> key /@ newConds, rest]
			]
		] /; FreeQ[ rc = rngToCondition[ rng], $Failed]
	]
	
inferenceRule[ RRPsplitForallGoal] = 
ps:PRFSIT$[ g:FML$[ _, u:Forall$TM[ rng:RNG$[ _, __], cond_, A_], __], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {vars = rngVariables[ rng], cl, cl2, newG, c, newGoal},
		cl = Switch[ cond, _And$TM, simplifiedAnd[ cond], _, cond];
		cl = Switch[ cl, _And$TM, List@@cl, _, {cl}];
		{cl2, newG} = extractPremises[ A];
		cl = DeleteCases[ Join[ cl, cl2], True];
		Scan[ (
				{c, cl} = RRPSplitAnd[ cl, Last[ vars]];
				vars = Most[ vars];
				newG = Forall$TM[ RNG$[ #], True, simplifiedImplies[ Implies$TM[ makeConjunction[ c, And$TM], newG]]]
			)&,
			Reverse[ rng]
		];
		newGoal = makeRRPFML[ formula -> newG, "marker" -> "G"];
		makeANDNODE[ makePRFINFO[ name -> RRPsplitForallGoal, used -> g],
			RRPToBeProved[ goal -> newGoal, kb -> k, rest]
		]
	]
]
	
(* 'RRPSplitAnd[ expr, v ]' is similar to 'splitAnd': It splits the list 'expr' into two (possibly empty) sublists 'a', 'b', where
	'a' contains all elements of 'expr' having 'v' as a free variable, and 'b' contains all other elements. *)
RRPSplitAnd[ expr_List, v:VAR$[ __]] :=
	Module[ {a, b, tmp},
		{tmp, {a, b}} = Reap[ Scan[
								With[ {fv = freeVariables[ #]},
									Sow[ #, If[ MemberQ[ fv, v], "a", "b"]]
								]&,
								expr
							],
							{"a", "b"}
						];
		If[ a =!= {}, a = First[ a]];
		If[ b =!= {}, b = First[ b]];
		{a, b}
	]
RRPSplitAnd[ expr_List, v:{___VAR$}] :=
	Module[ {a = {}, b = expr, tmp},
		Scan[ (
				{tmp, b} = RRPSplitAnd[ b, #];
				a = Join[ a, tmp]
			)&,
			v
		];
		{a, b}
	]
extractPremises[ Implies$TM[ prm_, cncl_]] :=
	Module[ {pl1, pl2, conclusion},
		{pl2, conclusion} = extractPremises[ cncl];
		pl1 = Switch[ prm, _And$TM, simplifiedAnd[ prm], _, prm];
		pl1 = Switch[ pl1, _And$TM, List@@pl1, _, {pl1}];
		{Join[ pl1, pl2], conclusion}
	]
extractPremises[ A_] := {{}, A}

inferenceRule[ RRPexistsKB] =
ps:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :>
	performProofStep[
		Module[ {r, c, f, fix, locC = getOptionalComponent[ ps, "constants"]},
			With[ {newAssms0 =
					Replace[
						k,
						FML$[ _?activatedQ, Exists$TM[ rng_, cond_, A_], __] :>
							With[ {rc = rngToCondition[ rng]},
								If[ FreeQ[ rc, $Failed],
									{{r, c, f}, fix} = arbitraryButFixed[ {rc, cond, A}, rng, {g, k, locC}];
									PrependTo[ locC, fix];
									{Map[ makeRRPFML[ formula -> #, "marker" -> "A"]&, DeleteCases[ Join[ r, {c, f}], True]], rngConstants[ fix]},
								(*else*)
									Null
								]
							],
						{1}
					]},
			With[ {simpPos = Position[ newAssms0, _List, {1}, Heads -> False]},
				If[ simpPos =!= {},
					With[ {newAssms1 = Extract[ newAssms0, simpPos], usedAssms = Extract[ k, simpPos]},
					With[ {simpAssms = newAssms1[[All, 1]]},
						makeANDNODE[
							makePRFINFO[ name -> RRPexistsKB, used -> usedAssms, generated -> simpAssms, "abf" -> newAssms1[[All, 2]]],
				        	RRPToBeProved[ goal -> g, kb -> joinNewKB[ Delete[ k, simpPos], Join @@ simpAssms], "constants" -> locC, "newAssms" -> key /@ (Join @@ simpAssms), rest]
						]
					]],
				(*else*)
					$Failed
				]
			]]
		]
	]

inferenceRule[ RIforallGoalSeqLeft] =
ps:PRFSIT$[ g:FML$[ _, Forall$TM[ RNG$[ _[ v:VAR$[ _SEQ0$|_SEQ1$], ___]], cond_, A_], __], k_List, _, rest___?OptionQ] :>
performProofStep[
	With[ {newG = RIforallGoalSeq[ g, Left]},
		If[ newG === $Failed,
			$Failed,
			makeANDNODE[ makePRFINFO[ name -> RIforallGoalSeqLeft, used -> g, generated -> newG, "var" -> v],
				RRPToBeProved[ goal -> newG, kb -> k, rest]
			]
		]
	]
]

RIforallGoalSeq[ g:FML$[ _, fa:(Forall$TM[ rng:RNG$[ r1:(_[ v:VAR$[ (SEQ0$|SEQ1$)[ n_Symbol]], ___])], cond_, A_]), __], mode:(Left|Right)] :=
	Module[ {rc = rngToCondition[ rng], g1, g2, newVar, subst, newCond, allVars, i},
		If[ FreeQ[ rc, $Failed],
			allVars = Replace[ DeleteDuplicates[ Cases[ Join[ rc, {cond, A}], VAR$[ x_Symbol] :> SymbolName[ x], Infinity]],
								{s_String?(StringMatchQ[ #, __ ~~ "$TM"]&) :> StringDrop[ s, -3], s_String?(StringMatchQ[ #, __ ~~ "$"]&) :> StringDrop[ s, -1]},
								{1}
						];
			newVar = Replace[ SymbolName[ n], {s_String?(StringMatchQ[ #, __ ~~ "$TM"]&) :> StringDrop[ s, -3], s_String?(StringMatchQ[ #, __ ~~ "$"]&) :> StringDrop[ s, -1]}];
			If[ MemberQ[ allVars, newVar],
				i = 0;
				While[ MemberQ[ allVars, newVar <> ToString[ i]], ++i];
				Begin[ "Theorema`Knowledge`"];
				newVar = Theorema`Language`VAR$[ ToExpression[ newVar <> ToString[ i] <> "$TM"]];
				End[],

				newVar = VAR$[ n]
			];
			g1 = Switch[ v,
					VAR$[ _SEQ0$],
					substituteFree[ simplifiedImplies[ Implies$TM[ simplifiedAnd[ makeConjunction[ Append[ rc, cond], And$TM]], A]], {v -> SEQ$[]}, "checkTypes" -> False],
					_,
					renameBound[ fa, {newVar}, "checkTypes" -> False, "postprocessing" -> Identity]
				];
			subst = Switch[ mode,
						Left,
						{v -> SEQ$[ newVar, v]},
						_,
						{v -> SEQ$[ v, newVar]}
					];
			newCond = simplifiedAnd[ makeConjunction[ Append[ rc, cond], And$TM]];
			g2 = Forall$TM[
					RNG$[ SIMPRNG$[ v], SIMPRNG$[ newVar]],
					substituteFree[ newCond, subst, "checkTypes" -> False, "postprocessing" -> Identity],
					Implies$TM[ simplifiedImplies[ Implies$TM[ newCond, A]], substituteFree[ A, subst, "checkTypes" -> False, "postprocessing" -> Identity]]
				];
			makeRRPFML[ formula -> And$TM[ g1, g2], "marker" -> "G"],

			$Failed
		]
	]
RIforallGoalSeq[ ___] := $Failed


(* ::Section::Closed:: *)
(* Rewriting *)


inferenceRule[ RRPgenerateRewriteRules] = 
this:PRFSIT$[ g_, k_List, _, opts___?OptionQ] :> performProofStep[
	Module[ { newKB = {}, newOpts, form, act, keys = {}},
		Block[ {$extractTagsForRewriting = True, $autoGenerateRules = True},
			(* We have to take into account all formulae that are not in the KB any more because they have already
			been transformed into rewrite rules. This is somewhat tedious, but probably there is no better way ... *)
			Scan[ 
				With[ {ky = key[ #]},
					If[ !MemberQ[ keys, ky],
						AppendTo[ keys, ky];
						With[ {new = simplifyFormulaByComputation[ #]},
							If[ new =!= $Failed, form = new]
						];
						newKB = addNewKB[ newKB, form]
					]
				]&,
				Join[ k, Map[ extractFormulaFromRule, Join @@ Cases[ this, _[ kbRules|goalRules|substRules|defRules, l_List] :> l]]]
			]
		];
		
		act = ruleActivity@this /. {h_[ RRPgenerateRewriteRules, _] :> h[ RRPgenerateRewriteRules, False]};
		
		newOpts = {opts} /. {h_[ r:(kbRules|goalRules|substRules|defRules), _] :> h[ r, {}],
							h_[ ruleActivity, _] :> h[ ruleActivity, act]};
		makeANDNODE[ makePRFINFO[ name -> RRPgenerateRewriteRules, used -> {}, generated -> {}], 
			RRPToBeProved[ goal -> g, kb -> newKB, Sequence @@ newOpts]]
	]
]

(* The following inference rule is almost completely the same as the one in "BasicTheoremaLanguageRules",
	with the following differences:
	- the condition on the goal being atomic is dropped
	- no book-keeping concerning rules that have already been tried is done
	- if no rewrite is possible, $Failed is returned
	- the goal is only rewritten to the "simplest" formulae (according to function "simplestFormulae")
*)
inferenceRule[ RRPgoalRewriting] = 
this:PRFSIT$[ g:FML$[ _, f_, __], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {usedSubsts, conds, newForms, simplifiedForms = {}, j, newNodes = {},
				rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "keys" -> Except[ key@g]]},
		newForms = replaceListAndTrack[ preprocessExpr[ f], rules];
		Switch[ newForms,
			{_List, _List, _List},
			j = 0;
			{newForms, usedSubsts, conds} =
				Catch[
					Scan[(
							++j;
							(* Goal rewriting should actually generate NO conditions. If a condition still appears, there must have gone something wrong. *)
							Assert[ newForms[[-1, j]]];
							With[ {tmp = followsImmediately[ #, rules, "RenameBound" -> False, "Depth" -> 2]},
								If[ MatchQ[ tmp, {_List, {}}],
									Throw[ {True, DeleteDuplicates[ Join[ newForms[[2, j]], First[ tmp]]], True}]
								]
							]
						)&,
						First[ newForms]
					];
					(* no need to postprocess conditions, since they have to be True anyway *)
					Throw[ MapAt[ postprocessExpr, newForms, 1]]
				],
			_,
			{newForms, usedSubsts, conds} = {{}, {}, {}}
		];
		If[ newForms === True,
			makeTERMINALNODE[ makePRFINFO[ name -> goalRewriting, used -> Prepend[ usedSubsts, g], generated -> {}], proved],
		(*else*)
			Do[
				AppendTo[ simplifiedForms, {makeRRPFML[ formula -> newForms[[j]], "marker" -> "G"], usedSubsts[[j]]}],
				{j, Length[ newForms]}
			];
			newNodes = Map[ makeANDNODE[
								makePRFINFO[ name -> goalRewriting, used -> Prepend[ Last[ #], g], generated -> First[ #]],
								RRPToBeProved[ goal -> First[ #], kb -> k, rest]
							]&,
							simplestFormulae[ simplifiedForms]
						];
			Switch[ Length[ newNodes],
				0,
				If[ TrueQ[ $applyAllPossibleInferences] && MatchQ[ f, _And$TM],
					Block[ {$proposedRules = {{RRPgoalRewriting}}},
						Replace[ this, {inferenceRule[ RRPandGoal], _ -> $Failed}]
					],
				(*else*)
					$Failed
				],
				1,
				First[ newNodes],
				_,
				If[ TrueQ[ $applyAllPossibleInferences],
					newNodes,
				(*else*)
					makeORNODE[ 
						makePRFINFO[ name -> multipleGoalRewriting, used -> used@newNodes, generated -> generated@newNodes],
						newNodes
					]
				]
			]
		]       	
	]
]

(* The following are the differences between "RRPelementarySubstitution" and "elementarySubstitution":
-	Rather than "replaceRepeatedAndTrack" the new function "RIReplace" is used
-	Hence, no conditions have to be checked afterwards, i.e. only one new proof situation will be generated
-	Also, "RRPelementarySubstitution" only applies rules with variables in their LHS. "RRPgroundSubstitution"
	applies all other rules (with ground LHS).
In fact, "RRPelementarySubstitution" and "RRPexpandDef" do almost exactly the same: The only difference is that
"RRPelementarySubstitution" uses (non-ground) elementary substitution rules and "RRPexpandDef" uses definition rules.
*)
inferenceRule[ RRPelementarySubstitution] = 
ps:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {rw = getOptionalComponent[ ps, "rewriting"]},
		Block[ {$forbiddenConditions = getOptionalComponent[ ps, "forbiddenConds"]},
			rewriteProofSit[
				g,
				k,
				getRuleClassesFromRW[ rw, "type" -> "Substitution", "ground" -> False],
				getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround],
				goalActivated[],
				activatedQ,
				False,
				RRPelementarySubstitution,
				rest
			]
		]
	]
]

inferenceRule[ RRPgroundSubstitution] = 
ps:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {rw = getOptionalComponent[ ps, "rewriting"]},
		Block[ {$forbiddenConditions = getOptionalComponent[ ps, "forbiddenConds"]},
			rewriteProofSit[
				g,
				k,
				getRuleClassesFromRW[ rw, "type" -> "Substitution", "ground" -> True],
				getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround],
				goalActivated[],
				activatedQ,
				False,
				RRPgroundSubstitution,
				rest
			]
		]
	]
]

(* The following are the differences between "RRPexpandDef" and "expandDef":
-	Rather than "replaceAllAndTrack" the new function "RIReplace" is used
-	Hence, no conditions have to be checked afterwards, i.e. only one new proof situation will be generated
*)
inferenceRule[ RRPexpandDef] = 
ps:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {rw = getOptionalComponent[ ps, "rewriting"]},
		Block[ {$forbiddenConditions = getOptionalComponent[ ps, "forbiddenConds"]},
			rewriteProofSit[
				g,
				k,
				getRuleClassesFromRW[ rw, "type" -> "ExplicitDefinition"],
				getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround],
				goalActivated[],
				activatedQ,
				False,
				RRPexpandDef,
				rest
			]
		]
	]
]

(*
	Remark:
	There is a difference between inference rules "RIexpandSuch" and "RRPimplicitDef": The former requires a suitable term to exist,
	whereas the latter doesn't. This is justified by the fact that if in, say,
	(1)		"forall x: f[x] := such y: P[x,y]"
	a suitable "y" does not exist (for some values of "x"), *the theory itself* is inconsistent since a formula like (1) merely abbreviates
	(1')	"forall x: P[x, f[x]]"
	Anything can be proved form inconsistent assumptions, hence we do not need to require terms to exist
	(note that inconsistencies can also be introduced by a single higher-order explicit definition!).
	In "RIexpandSuch" things are different: The implicit terms appearing there might not come from the theory, but rather might be instantiations
	of logical quantifiers. Thus, it is absolutely clear that existence of the respective terms *must* be required, for otherwise *every*
	existential goal would trivially be provable.
*)
inferenceRule[ RRPimplicitDef] = 
ps:PRFSIT$[ g_, K_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {rw = getOptionalComponent[ ps, "rewriting"], repl, gRules, implDefTerms, term, pos, activePos,
			const = {}, newK, ref, condRef, conds, out = {}, locC = getOptionalComponent[ ps, "constants"]}, 
		
		(* We only expand implicit definitions of symbols appearing in at least one activated formula (goal or KB),
			but we expand them everywhere (also in non-activated formulas). *)
		activePos = Select[ Range[ 1, Length[ K]], activatedQ[ key@K[[ #]]]&];
		
		repl = getRuleClassesFromRW[ rw, "type" -> "ImplicitDefinition"];
		gRules = getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround];
		Catch[
			Scan[ Function[ r,
					(* we only expand *one* alpha-equivalence class at a time, unless '$applyAllPossibleInferences' is set to 'True'. *)
					implDefTerms = getGroundInstances[ {formula@g, Map[ formula, K]}, r[[2, 1]], ExceptHeads -> {EqualDef$TM, IffDef$TM}];
					(* implDefTerms now is a list of the following form:
						'{{inst1, {p11, p12, ...}}, ...}'
						'insti' are ground instances of the pattern PREPROCESSED FOR H-O REWRITING!,
						and the 'pij' are the positions where alpha-equivalent expressions of 'insti' appear in '{g, K}'. *)
					If[ implDefTerms =!= {},
						Scan[ Function[ tp,
								{term, pos} = tp;
								If[ Catch[ Scan[ Switch[ First[ #], 1, If[ goalActivated[], Throw[ True]], 2, If[ MemberQ[ activePos, #[[2]]], Throw[ True]]]&, pos]; False],
									{const, newK, {ref, condRef}, conds} = makeImplDefSubst[ {term}, r[[2, 2]], Map[ formula, K], gRules, locC];
									Scan[ Function[ c,
											$rewriteRules = {};
											With[ {new = Catch[	(* catches all '$Failed' thrown by 'makeRRPFML' *)
															expandImplicitDefinition[
																c,
																pos,
																First[ r],
																ref,
																condRef,
																Map[ makeRRPFML[ formula -> #, "marker" -> "C"]&, conds],
																newK,
																term,
																g,
																K,
																locC,
																rest
															]
														]
													},
												If[ new =!= $Failed,
													AppendTo[ out, new];
													If[ !TrueQ[ $applyAllPossibleInferences], Throw[ Null]]
												]
											]
										],
										const
									]
								]
							],
							implDefTerms
						]
					]
				],
				repl
			]
		];
		Switch[ Length[ out],
			0,
			$Failed,
			1,
			First[ out],
			_,
			out
		]
	]
]

expandImplicitDefinition[ const_, p_List, usedKey_, ref_, condRef_List, conds_List, newK_List, term_, g_, K_List, locC_List, rest___] :=
	Module[ {subst, sPos, pos, newG, allGenerated, allUsed, i, newForm, substKB = {}, thinnedKB},
		subst = First[ rngConstants[ const]];
		sPos = Cases[ p, {1, r___} :> {r}];
		If[ sPos === {},
			(* no subst in goal *)
			newG = g;
			allGenerated = {{}};
			allUsed = {{}},
			(* new goal by substitution *)
			newG = makeRRPFML[ formula -> ReplacePart[ formula@g, sPos -> subst], "marker" -> "G"];
			allGenerated = {{newG}};
			allUsed = {{g}}
		];
		(* Do the same as for goal for all in KB *)
		pos = Replace[ GatherBy[ Cases[ p, {2, j_, r___} :> (j -> {r})], First], all:{j_ -> _List, ___} :> (j -> all[[All, 2]]), {1}];
		i = 1;
		Scan[ (
				If[ key@# =!= usedKey && ListQ[ sPos = Replace[ i, pos]], (* Don't rewrite the def itself. Def should not appear in KB anyway. *)
					newForm = makeRRPFML[ formula -> ReplacePart[ formula@#, sPos -> subst], "marker" -> "A"];
					AppendTo[ allGenerated, {newForm}];
					AppendTo[ allUsed, {#}];
					substKB = addNewKB[ substKB, newForm],
					
					Block[ {$autoGenerateRules = False}, substKB = addNewKB[ substKB, #]]
				];
				++i
			)&,
			K
		];
		(* add the characteristic properties to the KB and register them as newly generated *)
		PrependTo[ allGenerated, Map[ makeRRPFML[ formula -> #, "marker" -> "A"]&, newK]];
		(* remove all KB formulae where substitutions were performed *)
		thinnedKB = Delete[ K, Replace[ pos, (j_ -> _List) :> {j}, {1}]];
    	makeANDNODE[ makePRFINFO[ name -> RRPimplicitDef, used -> Prepend[ allUsed, {ref, Join[ condRef, conds]}], generated -> allGenerated,
    							  "introConstFor" -> {postprocessExpr[ term], subst}, "conditions" -> (conds =!= {})],
    		If[ conds === {},
    			RRPToBeProved[ goal -> newG, kb -> joinNewKB[ substKB, First[ allGenerated], Mode -> Prepend], "constants" -> Prepend[ locC, const], rest],
    			{
    				makeANDNODE[ makePRFINFO[ name -> Theorema`Provers`RRPcheckConditions, used -> {}, generated -> {conds}],
    					Map[ RRPToBeProved[ goal -> #, kb -> thinnedKB, rest]&, conds]
    				],
    				RRPToBeProved[ goal -> newG, kb -> joinNewKB[ substKB, Join[ First[ allGenerated], conds], Mode -> Prepend], "constants" -> Prepend[ locC, const], rest]
    			}
    		]
    	]
	]

Options[ getGroundInstances] = {ExceptHeads -> {}, Except -> {}};
getGroundInstances[ expr_, pattern_, opts___?OptionQ] :=
	Module[ {pp = preprocessExpr[ expr], ex, eh, allPos, out},
		{ex, eh} = {Except, ExceptHeads} /. {opts} /. Options[ getGroundInstances];
		eh = Alternatives@@Map[ Append[ #, BlankSequence[]]&, Position[ pp, (Alternatives@@eh)[ ___]]];
		allPos = DeleteCases[ Position[ pp, pattern], eh];
		out = Cases[ Transpose[ {Extract[ pp, allPos], allPos}],
				   {e_, _} /; (freeVariables[ e] === {} && !alphaMemberQ[ ex, e])
			];
		Replace[ Gather[ out, alphaEquivalent[ First[ #1], First[ #2]]&],
			all:{{inst_, _List}, ___List} :> {inst, all[[All, 2]]},
			{1}
		]
	]

makeImplDefSubst[ terms_List, def_RuleDelayed, K_List, gRules_List, locC_List] :=
	Module[ {newBody, abf, fix, allConst = {}, newK = {}, ref = {}, cond, condrefs = {}, allConds = {}, allCondRefs = {}, tmp},
		Scan[ (
				(* no preprocessing is necessary, since terms have already been preprocessed in 'getGroundInstances' *)
				{newBody, {ref, cond}} = Reap[ Replace[ #, def], {"ref", "cond"}];
				If[ ref =!= {},
					tmp = followsImmediately[ cond, gRules, "RenameBound" -> False, "Strict" -> !TrueQ[ $applyAllPossibleInferences]];
					If[ MatchQ[ tmp, {_List, _List}],
						{condrefs, cond} = tmp;
						{newBody, cond} = postprocessExpr[ {newBody, DeleteDuplicates[ cond]}];
						allConds = Join[ allConds, cond];
						allCondRefs = Join[ allCondRefs, condrefs];
						
						(* condition of the implicit definition is fulfilled for _all_ elements of this alpha-equivalence class *)
						{abf, fix} = arbitraryButFixed[ newBody[[2]], First[ newBody], {K, newK, locC}];
						(* newBody can only have the ONE variable from the such-quantifier as free variable, hence there is only ONE constant in fix *)
						AppendTo[ allConst, fix];
						AppendTo[ newK, abf]
					]
				]
			)&,
			terms
		];
		{allConst, newK, {Switch[ ref, {{_}}, ref[[1, 1]], _, ref], DeleteDuplicates[ allCondRefs]}, DeleteDuplicates[ allConds]}
	]
	
rewriteProofSit[ args___] :=
	If[ TrueQ[ $applyAllPossibleInferences],
		rewriteProofSitIndividually[ args],
		rewriteProofSitAllFormulas[ args]
	]

rewriteProofSitAllFormulas[ g_, k_List, repl_List, gRules_List, ga_, ka_, keepOrig_, n_Symbol, opts___?OptionQ] :=
	rewriteProofSitAllFormulas[ g, k, repl, gRules, First, True, True, ga, ka, keepOrig, n, opts]
rewriteProofSitAllFormulas[ _, _List, {}, _List, _, _, _, _, _, _, _Symbol, ___?OptionQ] := $Failed
rewriteProofSitAllFormulas[ g_, k_List, repl_List, gRules_List, mode_, cc_, pc_, ga_, ka_, keepOrig_, n_Symbol, opts___?OptionQ] :=
	Module[ {usedDef = {}, conds, allConds = {}, new, newG, newForm, newK = {}, rewritten = False, usedForms, genForms, replBy = {}, eh, l, fkey},
		If[ TrueQ[ ga],
			{new, usedDef, conds} = RIReplace[ gRules, Mode -> mode, CheckConditions -> cc, ProveConditions -> pc][
										preprocessExpr[ formula@g],
										repl[[All, 2]]
									];
		];
		
		Switch[ usedDef,
			{{__}, _List},
			{new, conds} = postprocessExpr[ {new, conds}];
			newG = makeRRPFML[ formula -> new, "marker" -> "G"];
			rewritten = True;
			(* We have to make 'FML$'s out of the conditions already here, because we need to add them to 'replBy',
				and 'replBy' must consist only of 'FML$'s. *)
			allConds = Map[ makeRRPFML[ formula -> #, "marker" -> "C"]&, DeleteDuplicates[ conds]];
			AppendTo[ replBy, {usedDef[[1, 1]], Join[ DeleteDuplicates[ Last[ usedDef]], allConds]}],
			_,
			newG = g;
			AppendTo[ replBy, {{}, {}}]
		];
		usedForms = {{g}};
		genForms = {{newG}};
		eh = If[ n === RRPexpandDef || TrueQ[ $applyAllPossibleInferences],
				{EqualDef$TM, IffDef$TM},
				{Equal$TM, Iff$TM, EqualDef$TM, IffDef$TM}
			];
		Scan[ Function[ fml,
				(* We do not want to rewrite
					- formulae into the boolean truth constant True (or a formula whose truth value is True),
					- formulae into formulae that are already in the KB, and
					- subexpressions of equalities and equivalences, unless $applyAllPossibleInferences is True.
				*)
				fkey = key@fml;
				If[ TrueQ[ ka[ fkey]],
		            {new, usedDef, conds} = RIReplace[ gRules, Mode -> mode, AcceptTrue -> False, Except -> Join[ k, newK],
		            									ExceptHeads -> eh, CheckConditions -> cc, ProveConditions -> pc]
		            						[
		            							preprocessExpr[ formula@fml],
		            							Cases[ repl, {Except[ fkey], r_, ___} :> r]
	            							],
	            	usedDef = {}
				];
	            Switch[ usedDef,
	            	{{__}, _List},
	            	(*If[ cc === False,
	            		(* In that case, usedDef[[2]] is still empty *)
						{usedDef[[2]], conds} = followsImmediately[ conds, gRules, "RenameBound" -> False, "Strict" -> False]
					];*)
	            	{new, conds} = postprocessExpr[ {new, conds}];
	        		newForm = makeRRPFML[ formula -> new, "marker" -> "A"];
	        		newK = addNewKB[ newK, newForm, Mode -> Append];
	        		If[ TrueQ[ keepOrig], Block[ {$autoGenerateRules = False}, newK = addNewKB[ newK, fml, Mode -> Append]]];
	                AppendTo[ usedForms, {fml}];
	                AppendTo[ genForms, {newForm}];
	                conds = Map[
	                			If[ (l = Cases[ allConds, _[ _, #, ___], {1}, 1]) =!= {},
	                				First[ l],
	                				l = makeRRPFML[ formula -> #, "marker" -> "C"];
	                				AppendTo[ allConds, l];
	                				l
	                			]&,
	                			DeleteDuplicates[ conds]
	                		];
					AppendTo[ replBy, {usedDef[[1, 1]], Join[ DeleteDuplicates[ Last[ usedDef]], conds]}];
	                rewritten = True,
	                _,
	                (* else: no rewrite in this formula *)
	                Block[ {$autoGenerateRules = False}, newK = addNewKB[ newK, fml, Mode -> Append]] (* rewrite rules from this formula are already there *)
	            ]
			],
            k
        ];
        If[ rewritten,
        	If[ ListQ[ $forbiddenConditions], $forbiddenConditions = Join[ Map[ formula, allConds], $forbiddenConditions]];
        	(* allConds = Replace[ GatherBy[ allConds, First], all:{{c_, _}..} :> {makeRRPFML[ formula -> c, "marker" -> "C"], DeleteCases[ all[[All, 2]], 0]}, {1}]; *)
        	makeANDNODE[ makePRFINFO[ name -> n, used -> usedForms, generated -> genForms, "usedForms" -> replBy, "conditions" -> (allConds =!= {})],
        		If[ allConds === {},
					RRPToBeProved[ goal -> newG, kb -> newK, opts],
				(*else*)
					{
						Block[ {$rewriteRules = {}},
							makeANDNODE[ makePRFINFO[ name -> Theorema`Provers`RRPcheckConditions, used -> {}, generated -> {allConds}],
								Map[ RRPToBeProved[ goal -> #, kb -> k, "forbiddenConds" -> $forbiddenConditions, opts]&, allConds]
							]
						],
						RRPToBeProved[ goal -> newG, kb -> joinNewKB[ newK, allConds], opts]
					}
				]
        	],
        	
			$Failed
        ]
	]
	
rewriteProofSitIndividually[ g_, k_List, repl_List, gRules_List, ga_, ka_, keepOrig_, n_Symbol, opts___?OptionQ] :=
	rewriteProofSitIndividually[ g, k, repl, gRules, True, True, ga, ka, keepOrig, n, opts]
rewriteProofSitIndividually[ _, _List, {}, _List, _, _, _, _, _, _Symbol, ___?OptionQ] := {}
rewriteProofSitIndividually[ g_, k_List, repl_List, gRules_List, cc_, pc_, ga_, ka_, keepOrig_, n_Symbol, opts___?OptionQ] :=
	Module[ {out = {}, possibilities, usedDef, conds, new, newF, thinnedK, newK, replBy = {}, eh, fkey},
		If[ TrueQ[ ga],
			(* Attention! If definedness definitions are used, 'RIReplace' is probably not the right method.
				This is not only because of 'exceptHeads', but also because expanding definedness definitions in the goal
				heavily relies on the fact that definitions are expanded one after the other ($definednessExpressions). *)
			possibilities = RIReplace[ gRules, Mode -> List, CheckConditions -> cc, ProveConditions -> pc][
						preprocessExpr[ formula@g],
						repl[[All, 2]]
					],
					
			possibilities = {}
		];
		Scan[ Function[ t,
				{new, usedDef, conds} = t;
				(*{refs, conds} = followsImmediately[ conds, gRules, "RenameBound" -> False, "Strict" -> False];*)
				{new, conds} = postprocessExpr[ {new, conds}];
				newF = makeRRPFML[ formula -> new, "marker" -> "G"];
				$rewriteRules = {};
				If[ conds === {},
					replBy = {{usedDef[[1, 1]], DeleteDuplicates[ Last[ usedDef]]}};
					AppendTo[ out, makeANDNODE[ makePRFINFO[ name -> n, used -> {{g}}, generated -> {{newF}}, "usedForms" -> replBy],
										RRPToBeProved[ goal -> newF, kb -> k, opts]
									]
					],
					
					conds = makeRRPFML[ formula -> And$TM@@DeleteDuplicates[ conds], "marker" -> "C"];
					replBy = {{usedDef[[1, 1]], Append[ DeleteDuplicates[ Last[ usedDef]], conds]}};
					AppendTo[ out, makeANDNODE[ makePRFINFO[ name -> n, used -> {{g}}, generated -> {{newF}}, "usedForms" -> replBy, "conditions" -> conds],
										{
											Block[ {$rewriteRules = {}}, RRPToBeProved[ goal -> conds, kb -> k, opts]],
											RRPToBeProved[ goal -> newF, kb -> addNewKB[ k, conds], opts]
										}
									]
					]
				]
			],
			possibilities
		];
		
		eh = If[ n === RRPexpandDef || TrueQ[ $applyAllPossibleInferences],
				{EqualDef$TM, IffDef$TM},
				{Equal$TM, Iff$TM, EqualDef$TM, IffDef$TM}
			];
		Scan[ Function[ fml,
				(* We do not want to rewrite
					- formulae into the boolean truth constant True (or a formula whose truth value is True),
					- formulae into formulae that are already in the KB, and
					- subexpressions of equalities and equivalences, unless $applyAllPossibleInferences is True.
				*)
				fkey = key@fml;
				If[ TrueQ[ ka[ fkey]],
					possibilities = RIReplace[ gRules, Mode -> List, CheckConditions -> cc, ProveConditions -> pc,
												ExceptHeads -> eh, Except -> k, AcceptTrue -> False]
										[
												preprocessExpr[ formula@fml],
												Cases[ repl, {Except[ fkey], r_, ___} :> r]
										],
						
	            	possibilities = {}
				];
				If[ possibilities =!= {},
					thinnedK = If[ TrueQ[ keepOrig], k, DeleteCases[ k, FML$[ key@fml, ___]]];
					Scan[ Function[ t,
							{new, usedDef, conds} = t;
							(*{refs, conds} = followsImmediately[ conds, gRules, "RenameBound" -> False, "Strict" -> False];*)
							{new, conds} = postprocessExpr[ {new, conds}];
							newF = makeRRPFML[ formula -> new, "marker" -> "A"];
							$rewriteRules = {};
							newK = addNewKB[ thinnedK, newF, Mode -> Prepend];
							If[ conds === {},
								replBy = {{{}, {}}, {usedDef[[1, 1]], DeleteDuplicates[ Last[ usedDef]]}};
								AppendTo[ out,
										makeANDNODE[ makePRFINFO[ name -> n, used -> {{}, {fml}}, generated -> {{}, {newF}}, "usedForms" -> replBy],
											RRPToBeProved[ goal -> g, kb -> newK, opts]
										]
								],
								
								conds = makeRRPFML[ formula -> And$TM@@DeleteDuplicates[ conds], "marker" -> "C"];
								replBy = {{{}, {}}, {usedDef[[1, 1]], Append[ DeleteDuplicates[ Last[ usedDef]], conds]}};
								AppendTo[ out,
										makeANDNODE[ makePRFINFO[ name -> n, used -> {{}, {fml}}, generated -> {{}, {newF}}, "usedForms" -> replBy, "conditions" -> conds],
											{
												Block[ {$rewriteRules = {}}, RRPToBeProved[ goal -> conds, kb -> k, opts]],
												RRPToBeProved[ goal -> g, kb -> addNewKB[ newK, conds], opts]
											}
										]
								]
							]
						],
						possibilities
					]
				]
			],
            k
        ];
        out
	]
	
	


(* ::Section::Closed:: *)
(* Interactive Rules *)


inferenceRule[ RRPorGoalInteractive] = 
PRFSIT$[ g:FML$[ _, o:Or$TM[ a_, b__], lab_, ___] /; FreeQ[ o, _META$], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {newGoals, negAssm, newG, sel},
		newGoals = MapIndexed[ makeRRPFML[ formula -> #1, label -> lab <> "." <> ToString[ #2[[1]]], "marker" -> "G"]&, {a, b}];
		sel = getInferenceRuleParams[ {1}, selectGoalDialog[ newGoals, k, "selectGoal"]];	(* '{1}' is the list of default values, not the parameter-index! *)
		If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ newGoals],
			newG = newGoals[[sel]];
			negAssm = DeleteCases[ MapIndexed[ If[ #2 === {sel},
											$Failed,
											makeRRPFML[ formula -> sinkNegation[ Not$TM[ #1]], label -> lab <> "." <> ToString[ #2[[1]]], "marker" -> "A"]
										]&, {a, b}
									],
					  				$Failed
					  ];
			makeANDNODE[ makePRFINFO[ name -> orGoal, used -> g, generated -> {Append[ negAssm, newG]}],
				RRPToBeProved[ goal -> newG, kb -> joinNewKB[ k, negAssm, Mode -> Prepend], "newAssms" -> key /@ negAssm, rest]
			],
		(*else*)
			$Failed
		]
	]
]

inferenceRule[ RRPexchangeGoalKBInteractive] = 
PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {negGoal, negFml, fml, sel = {}},
		If[ ListQ[ $knowledgeActivated] && Length[ $knowledgeActivated] === 1,
			sel = Position[ k, FML$[ First[ $knowledgeActivated], __], {1}, 1, Heads -> False]
		];
		If[ sel === {},
			If[ Theorema`Provers`$allowInteraction =!= False, sel = selectFormulaDialog[ g, k, "exchangeGoal"]],
		(*else*)
			sel = sel[[1, 1]]
		];
		If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ k],
			negGoal = makeRRPFML[ formula -> sinkNegation[ Not$TM[ formula@g]], "marker" -> "A"];
			fml = k[[sel]];
			negFml = makeRRPFML[ formula -> sinkNegation[ Not$TM[ formula@fml]], "marker" -> "G"];
			makeANDNODE[ makePRFINFO[ name -> RRPexchangeGoalKBInteractive, used -> {g, fml}, generated -> {negGoal, negFml}], 
					RRPToBeProved[
						goal -> negFml,
						kb -> addNewKB[ Delete[ k, sel], negGoal, Mode -> Prepend],
						"newAssms" -> {key@negGoal},
						rest
                	]
            ],
        (*else*)
			$Failed
		]
	]
]

inferenceRule[ RRPsplitForallGoalInteractive] =
ps:PRFSIT$[ g:FML$[ _, u:Forall$TM[ rng:RNG$[ _, __], cond_, A_], __], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {vars, cl, cl2, newG, c, newGoal, num, simplify},
		num = getInferenceRuleParams[ 2, {1, True}, splitForallDialog[ g, k]];
		If[ MatchQ[ num, {_Integer?(1 <= # < Length[ rng] &), _}],
			{num, simplify} = num;
			vars = rngVariables[ Drop[ rng, num]];
			If[ TrueQ[ simplify],
				cl = Switch[ cond, _And$TM, simplifiedAnd[ cond], _, cond];
				cl = Switch[ cl, _And$TM, List@@cl, _, {cl}];
				{cl2, newG} = extractPremises[ A];
				cl = DeleteCases[ Join[ cl, cl2], True];
				{cl, c} = RRPSplitAnd[ cl, vars];
				newG = simplifiedImplies[ Implies$TM[ makeConjunction[ cl, And$TM], newG]];
				cl = {};
				cl2 = {},
				
				newG = A;
				cl = Switch[ cond, _And$TM, List@@cond, _, {cond}];
				{cl2, cl} = RRPSplitAnd[ cl, vars];
				c = {}
			];
			newG = Forall$TM[ Take[ rng, num],
								makeConjunction[ cl, And$TM],
								simplifiedImplies[ Implies$TM[ makeConjunction[ c, And$TM],
										Forall$TM[ Drop[ rng, num], makeConjunction[ cl2, And$TM], newG]
								]]
					];
			newGoal = makeRRPFML[ formula -> newG, "marker" -> "G"];
			makeANDNODE[ makePRFINFO[ name -> RRPsplitForallGoal, used -> g],
				RRPToBeProved[ goal -> newGoal, kb -> k, rest]
			],
		(*else*)
			$Failed
		]
	]
]

inferenceRule[ RRPinstantiateInteractive] = 
ps:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
    Module[ {fml, rc, instVars, instRng, thinnedRng, fInst, fCond, sel = {}, rng, cond, A, subst, domPos, checkConds, keepOrig, newK, rw = getOptionalComponent[ ps, "rewriting"]},
    	If[ goalActivated[] && MatchQ[ g, FML$[ _, Exists$TM[ _, _, _], ___]],
    		sel = 0,
    	(*else*)
	    	If[ ListQ[ $knowledgeActivated],
	    		sel = Position[ k, FML$[ Alternatives@@$knowledgeActivated, Forall$TM[ _, _, _], __], {1}, Heads -> False]
	    	]
    	];
    	If[ sel === {},
	    	If[ Theorema`Provers`$allowInteraction =!= False, sel = selectFormulaDialog[ g, k, FML$[ _, Exists$TM[ _, _, _], __], FML$[ _, Forall$TM[ _, _, _], __], "selectFormula"]];
	    	If[ IntegerQ[ sel] && sel > 0, sel = {{sel}}];
    	];
    	Switch[ sel,
    		0,
    		(* Instantiate goal *)
    		(* Note the 's' in 'instantiateExist_s_GoalInteractive', making the function different from the public one!*)
    		rng = First[ formula@g];
	        subst = getInferenceRuleParams[ All, {}, instantiateExistsGoalInteractive[ g, getOptionalComponent[ ps, "constants"] /. RNG$ -> Sequence, k]];
	        If[ MatchQ[ subst, {__Expression}],
				subst = makeSubstitution[ rngVariables[ rng], First /@ subst];
				If[ MatchQ[ subst, {__}],
		            instRng = Select[ rng, MemberQ[ First /@ subst, First[ #]]&];
					thinnedRng = Complement[ rng, instRng];
		            rc = rngToCondition[ instRng];
		            If[ !FreeQ[ rc, $Failed],
						$Failed,
					(*else*)
						cond = (formula@g)[[2]];
			        	A = Last[ formula@g];
			        	fInst = makeExist[ thinnedRng, rc, cond, A, subst];
			        	If[ fInst === $Failed,
			        		$Failed,
			        	(*else*)
			            	fInst = makeRRPFML[ formula -> fInst, "marker" -> "G"];
			            	makeANDNODE[ makePRFINFO[ name -> existsGoalInteractive, used -> g, "instantiation" -> subst],
			                	RRPToBeProved[ goal -> fInst, kb -> k, rest]]
			        	]
		            ],
				(*else*)
					$Failed
				],
			(*else*)
				$Failed
	        ],
    		
    		{__List},
    		(* Instantiate assumptions *)
        	fml = Extract[ k, sel];
        	If[ MatchQ[ fml, {FML$[ _, Forall$TM[ _, _, _], __]..}],
	        	rng = Composition[ First, formula] /@ fml;
	        	rng = Select[ First[ rng], Function[ r, Catch[ Scan[ If[ !MemberQ[ #, r], Throw[ False]]&, Rest[ rng]]; True]]];
	            subst =
	            	getInferenceRuleParams[
	            		All,
	            		{{True, True|False}, {True, True|False}},
	            		instantiateForallKBInteractive[ getOptionalComponent[ ps, "constants"] /. RNG$ -> Sequence, rng, {g, Join[ fml, Delete[ k, sel]]}, Alternatives@@(key/@fml)]
	            	];
	            If[ MatchQ[ subst, {__Expression, _, _}],
	            	checkConds = subst[[-2]];
	            	keepOrig = Last[ subst];
	                subst = makeSubstitution[ rngVariables[ rng], First /@ Drop[ subst, -2]];
	                If[ MatchQ[ subst, {__}],
		                instVars = First/@subst;
			        	instRng = Select[ rng, MemberQ[ instVars, First[ #]]&];
			            rc = rngToCondition[ instRng];
			            If[ FreeQ[ rc, $Failed],
			            	If[ TrueQ[ checkConds],
								domPos = Position[ rc, (DomainOperation$TM[ _?((freeVariables[ #] === {}) &), Element$TM][ _?(MemberQ[ instVars, #]&)]), {1}, Heads -> False];
								checkConds = substituteFree[ Extract[ rc, domPos], subst];
								If[ checkConds === $Failed, Throw[ $Failed]];	(* caught in 'performProofStep' *)
								rc = Delete[ rc, domPos],
							(*else*)
								checkConds = {}
							];
							A = Replace[ fml, FML$[ _, Forall$TM[ r_RNG$, c_, e_], __] :> substituteFree[ {Complement[ r, instRng], c, e}, subst], {1}];
							If[ MemberQ[ A, $Failed], Throw[ $Failed]];
							rc = DeleteCases[ substituteFree[ rc, subst], True];
							If[ MemberQ[ rc, $Failed], Throw[ $Failed]];
							fInst = makeRRPFML[ formula -> makeForall[ rc, A], "marker" -> "A"];
							If[ TrueQ[ keepOrig],
								newK = k,
							(*else*)
								newK = Delete[ k, sel];
								rw = removeRulesFromRW[ rw, "keys" -> Alternatives @@ (key /@ fml)];
							];
			            	If[ checkConds === {},
				                makeANDNODE[ makePRFINFO[ name -> RRPforallKBInteractive, used -> fml, generated -> {{fInst}}, "instantiation" -> subst],
				                    RRPToBeProved[ goal -> g, kb -> addNewKB[ newK, fInst, Mode -> Prepend], "rewriting" -> rw, "newAssms" -> {key@fInst}, rest]
				                ],
				            (*else*)
				                fCond = makeRRPFML[ formula -> makeConjunction[ checkConds, And$TM], "marker" -> "C"];
				                makeANDNODE[ makePRFINFO[ name -> RRPforallKBInteractive, used -> fml, generated -> {{fInst, fCond}}, "instantiation" -> subst],
				                	{
				                		Block[ {$proposedRules = {{RRPandGoal}}}, RRPToBeProved[ goal -> fCond, kb -> k, rest]],
				                    	RRPToBeProved[ goal -> g, kb -> joinNewKB[ newK, {fInst, fCond}, Mode -> Prepend], "rewriting" -> rw, "newAssms" -> {key@fInst, key@fCond}, rest]
				                	}
				                ]
			            	],
			            (*else*)
			                $Failed
			            ],
					(*else*)
						$Failed
					],
				(*else*)
	            	$Failed
	            ],
	        (*else*)
	            $Failed
	        ],
	        
	        _,
	        $Failed
    	]
    ]
]

makeForall[ cond_List, A:{{_, _, _}..}] :=
	Module[ {conds, commonConds = {}, cnew, enew, p},
		conds = Replace[ A, {r_, c_, e_} :>
								With[ {c0 = Switch[ c, _And$TM, List@@c, True, {}, _, {c}]},
									{cnew, enew} = Switch[ e,
														_Implies$TM,
														With[ {si = simplifiedImplies[ e]},
															Switch[ si,
																_Implies$TM,
																Switch[ First[ si],
																	_And$TM,
																	{Join[ c0, List@@First[ si]], Last[ si]},
																	_,
																	{Append[ c0, First[ si]], Last[ si]}
																],
																_,
																{c0, e}
															]
														],
														_,
														{c0, e}
													];
									cnew = DeleteDuplicates[ cnew];
									p = Position[ cnew, _?((freeVariables[ #] === {}) &), {1}, Heads -> False];
									{Extract[ cnew, p], r, Delete[ cnew, p], enew}
								],
							{1}
				];
		commonConds = Intersection@@conds[[All, 1]];
		conds = Replace[ conds, {c1_List, r_RNG$, c2_List, e_} :>
									With[ {comp = Complement[ c1, commonConds]},
										simplifiedImplies[ Implies$TM[ makeConjunction[ comp, And$TM], makeForall[ r, c2, e]]]
									], {1}
				];
		simplifiedImplies[ Implies$TM[ makeConjunction[ Join[ cond, commonConds], And$TM], makeConjunction[ conds, And$TM]]]
	]
makeForall[ r:RNG$[ __], cond_List, A_] :=
	With[ {p = Position[ cond, _?((freeVariables[ #] === {}) &), {1}, Heads -> False]},
		If[ p === {},
			Forall$TM[ r, True, makeForall[ RNG$[], cond, A]],
			Implies$TM[ makeConjunction[ Extract[ cond, p], And$TM], Forall$TM[ r, True, makeForall[ RNG$[], Delete[ cond, p], A]]]
		]
	]
makeForall[ RNG$[], {}, A_] := A
makeForall[ RNG$[], cond:{__}, A_] := Implies$TM[ makeConjunction[ cond, And$TM], A]

makeExist[ RNG$[], cond1_List, cond2_, A_, subst_] :=
	With[ { c = substituteFree[ DeleteCases[ Join[ cond1, {cond2, A}], True], subst]},
		Switch[ c,
			$Failed,
			$Failed,
			
			{},
			True,
			
			_,
			And$TM @@ c
		]
	]
makeExist[ r:RNG$[__], cond1_List, cond2_, A_, subst_] :=
	With[ {out = substituteFree[ r, subst]},
		If[ out === $Failed,
			$Failed,
			Exists$TM[ out, True, makeExist[ RNG$[], cond1, cond2, A, subst]]
		]
	]
	
inferenceRule[ RRPselectImplicationInteractive] = 
this:PRFSIT$[ g_FML$, k_List?(MemberQ[ #, FML$[ _, Implies$TM[ _, _], ___]]&), _, rest___?OptionQ] :> performProofStep[
	Module[ {sel = {}, fml, assm, concl, assmGoal, provePremise = getInferenceRuleParams[ {True}, True], rw = getOptionalComponent[ this, "rewriting"]},
		If[ ListQ[ $knowledgeActivated] && Length[ $knowledgeActivated] === 1,
    		sel = Position[ k, FML$[ First[ $knowledgeActivated], __], {1}, 1, Heads -> False];
    		If[ sel =!= {},
    			sel = sel[[1, 1]];
    			If[ !MatchQ[ k[[sel]], FML$[ _, Implies$TM[ _, _], ___]], sel = {}]
    		]
    	];
    	If[ sel === {} && Theorema`Provers`$allowInteraction =!= False,
			sel = selectFormulaDialog[ g, k, Null, FML$[ _, Implies$TM[ _, _], ___], "selectImplication", "DialogOneOption" -> True,
					"AdditionalMenuItems" -> {Tooltip[ Toggler[ Dynamic[ provePremise], {True -> Text[ translate[ "provePremise"]], False -> Text[ translate[ "distinguishCases"]]}, Appearance -> "Frameless"], translate[ "ttProveDistinguish"]]}
				]
    	];
		If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ k],
			fml = k[[sel]];
			rw = removeRulesFromRW[ rw, "keys" -> key@fml];	(* The old rewrite rules are not needed any more. *)
			assm = makeRRPFML[ formula -> First[ formula@fml], "marker" -> "A"];
			concl = makeRRPFML[ formula -> Last[ formula@fml], "marker" -> "A"];
			If[ TrueQ[ provePremise],
				assmGoal = makeRRPFML[ formula -> First[ formula@fml], "marker" -> "G"];
				makeANDNODE[ makePRFINFO[ name -> RRPselectImplicationInteractive, used -> {fml, g}, generated -> {{assmGoal}, {assm, concl}}], 
	                	{
	                		RRPToBeProved[ goal -> assmGoal, kb -> k, rest],
	                	 	RRPToBeProved[ goal -> g, kb -> joinNewKB[ Delete[ k, sel], {assm, concl}, Mode -> Prepend], "rewriting" -> rw, "newAssms" -> {key@assm, key@concl}, rest]
	                	}
	            ],
			(*else*)
	            assmGoal = makeRRPFML[ formula -> sinkNegation[ Not$TM[ First[ formula@fml]]], "marker" -> "A"];
				makeANDNODE[ makePRFINFO[ name -> RRPselectImplicationInteractive, used -> {fml, g}, generated -> {{assm, concl}, {assmGoal}}], 
	                	{
	                		RRPToBeProved[ goal -> g, kb -> addNewKB[ k, assmGoal, Mode -> Prepend], "newAssms" -> {key@assmGoal}, rest],
	                		Block[ {$rewriteRules = {}},
	                			RRPToBeProved[ goal -> g, kb -> joinNewKB[ Delete[ k, sel], {assm, concl}, Mode -> Prepend], "rewriting" -> rw, "newAssms" -> {key@assm, key@concl}, rest]
	                		]
	                	}
	            ]
			],
		(*else*)
			$Failed
		]
	]
]
	
inferenceRule[ RRPrewritePS] = 
this:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {rw = getOptionalComponent[ this, "rewriting"], fa, ga, dr, substRules, defRules, gRules, mode, cc, pc, kOrig},
		dr = getInferenceRuleParams[
				All,
				{{2, "first"|"all"|"repeated"|"individual"}, {True, True|False}, {True, True|False}, {False, True|False}},
				rewritingDialog[ g, k, rw, {"Substitution" -> translate[ "substRules"], "ExplicitDefinition" -> translate[ "defRules"]}, Hold[ fa]]
			];
		Switch[ dr,
			{{_List, _List}, _, _, _, _},
			{{substRules, defRules}, mode, cc, pc, kOrig} = dr;
			ga = fa[ key@g],
			
			{__, _, _, _, _},
			{substRules, defRules} = getRewriteRules[ Drop[ dr, -4], k, getOptionalComponent[ this, "newAssms"], rw, {"Substitution", "ExplicitDefinition"}];
			{mode, cc, pc, kOrig} = Take[ dr, -4];
			ga = goalActivated[];
			fa = activatedQ,
			
			_,
			mode = $Failed
		];
		If[ mode =!= $Failed,
			gRules = getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround];
			If[ StringQ[ mode], mode = Replace[ mode, {"first" -> 1, "all" -> 2, "repeated" -> 3, "individual" -> 4}]];
			If[ mode === 4,
				Which[
					defRules === {},
					rewriteProofSitIndividually[ g, k, substRules, gRules, cc, pc, ga, fa, kOrig, RRPelementarySubstitution, rest],
					substRules === {},
					rewriteProofSitIndividually[ g, k, defRules, gRules, cc, pc, ga, fa, kOrig, RRPexpandDef, rest],
					True,
					$Failed
				],
				
				mode = Switch[ mode,
						2,
						All,
						3,
						Repeated,
						_,
						First
					];
				(* Attention! Definitions and substitutions cannot be applied together. *)
				DeleteCases[
					{
						rewriteProofSitAllFormulas[ g, k, substRules, gRules, mode, cc, pc, ga, fa, kOrig, RRPelementarySubstitution, rest],
						rewriteProofSitAllFormulas[ g, k, defRules, gRules, mode, cc, pc, ga, fa, kOrig, RRPexpandDef, rest]
					},
					$Failed
				]
			],
		(*else*)
			$Failed
		]
	]
]

getRewriteRules[ ids_List, k_List, newAssms_List, rw_List, ruleTypes_List] :=
	With[ {fmlPos = Position[ ids, String[ _String]|_Expression, {1}, Heads -> False]},
	With[ {keys =
				Alternatives @@ Cases[
					Replace[
						Delete[ ids, fmlPos],
						{
							"this"|{"this"} :> Sequence @@ newAssms,
							{"this", i_Integer} :>
							Quiet[ Check[ newAssms[[i]], Sow[ {"Warning", "New assumption " <> ToString[ i] <> " does not exist."}, "Message"]; Null]],
							{"this", i_Integer, j_Integer} :>
							Quiet[ Check[ Sequence @@ newAssms[[i;;j]], Sow[ {"Warning", "New assumptions " <> ToString[ i] <> " through " <> ToString[ j] <> " do not exist."}, "Message"]; Null]],
							j_Integer :>
							Quiet[ Check[ key@(k[[j]]), Sow[ {"Warning", "Assumption " <> ToString[ j] <> " does not exist."}, "Message"]; Null]]
						},
						{1}
					],
					_List
				],
			patts =
				Alternatives @@ Replace[
					Extract[ ids, fmlPos],
					{
						String[ l_String] :> FML$[ _, _, l, ___],
						Expression[ expr_] :> FML$[ _, _?(alphaEquivalent[ expr, #] &), __]
					},
					{1}
				]},
	With[ {out =
			If[ patts === Alternatives[],
				Table[ {}, {Length[ ruleTypes]}],
			(*else*)
				Map[ Function[ {tp}, Select[ getRuleClassesFromRW[ rw, "type" -> tp], MatchQ[ extractFormulaFromRule[ #], patts] &]], ruleTypes]
			]},
		MapThread[ Function[ {rules, tp}, Join[ rules, getRuleClassesFromRW[ rw, "type" -> tp, "keys" -> keys]]], {out, ruleTypes}]
	]]];

inferenceRule[ RRPintroduceFormula] = 
this:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {expr, cd, repl, b1, b2 = $Failed, fml, fml2 = Sequence[], newK, rules, rw = getOptionalComponent[ this, "rewriting"]},
		expr = getInferenceRuleParams[ 3, {False, False}, formulaInputDialog[ g, k, getOptionalComponent[ this, "constants"] /. RNG$ -> Sequence, WindowTitle -> translate[ "introduceFormula"]]];
		If[ MatchQ[ expr, {Expression[ Except[ Null|$Failed]], _, _}],
			{expr, cd, repl} = expr;
			expr = First[ expr];
			If[ isSingleExpression[ expr],
				If[ TrueQ[ cd],
					fml = makeRRPFML[ formula -> expr, "marker" -> "A"];
					fml2 = makeRRPFML[ formula -> Not$TM[ expr], label -> label@fml <> ".\[Not]", "marker" -> "A"];
					b1 = RRPToBeProved[ goal -> g, kb -> addNewKB[ k, fml2, Mode -> Prepend], "newAssms" -> {key@fml2}, rest],
				(*else*)
					fml = makeRRPFML[ formula -> expr, "marker" -> "C"];
					b1 = RRPToBeProved[ goal -> fml, kb -> k, rest]
				];
				Block[ {$rewriteRules = {}},
					newK = addNewKB[ k, fml, Mode -> Prepend];
					If[ TrueQ[ repl] && Length[ $rewriteRules] === 1 && Length[ rules = getRuleClassesFromRW[ First[ $rewriteRules], "type" -> "Substitution"]] === 1,
						Block[ {$forbiddenConditions = getOptionalComponent[ this, "forbiddenConds"]},
							b2 = rewriteProofSitAllFormulas[ g, newK, rules, getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround],
																All, True, False, True, Function[ True], False, RRPelementarySubstitution, "newAssms" -> {key@fml}, rest]
						]
					];
					If[ b2 === $Failed, b2 = RRPToBeProved[ goal -> g, kb -> newK, "newAssms" -> {key@fml}, rest]]
				];
				makeANDNODE[ makePRFINFO[ name -> RRPintroduceFormula, used -> {}, generated -> {fml, fml2}], {b1, b2}],
			(*else*)
				$Failed
			],
		(*else*)
			$Failed
		]
	]
]

inferenceRule[ RRPaddAbbreviation] = 
this:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {sym, expr, equal, i, fml},
		expr = getInferenceRuleParams[ 3, {True}, formulaAbbrevDialog[ g, k, getOptionalComponent[ this, "constants"] /. RNG$ -> Sequence, WindowTitle -> translate[ "addAbbreviation"]]];
		If[ MatchQ[ expr, {Expression[ Except[ Null|$Failed]], Expression[ Except[ Null|$Failed]], _}],
			{sym, expr, equal} = expr;
			sym = First[ sym];
			expr = First[ expr];
			If[ MatchQ[ sym, _Symbol] && isSingleExpression[ expr],
				i = Max[ Cases[ {g, k}, FIX$[ sym, n_] -> n, Infinity]];
				i = If[ i < 0, 0, i + 1];
				sym = FIX$[ sym, i];
				fml = makeRRPFML[ formula -> If[ TrueQ[ equal], EqualDef$TM[ sym, expr], IffDef$TM[ sym, expr]], "marker" -> "Abbrv"];
				makeANDNODE[ makePRFINFO[ name -> RRPaddAbbreviation, used -> {}, generated -> {fml}],
					RRPToBeProved[
						goal -> g,
						kb -> addNewKB[ k, fml, Mode -> Append],
						"constants" -> Prepend[ getOptionalComponent[ this, "constants"], RNG$[ SIMPRNG$[ sym]]], 
						"newAssms" -> {key@fml},
						rest
					]
				],
			(*else*)
				$Failed
			],
		(*else*)
			$Failed
		]
	]
]


(* ::Section::Closed:: *)
(* Proof Management *)


inferenceRule[ RRPaddKnowledge] = 
PRFSIT$[ g_, k_List, pid_, rest___?OptionQ] :> performProofStep[
	Module[ {dr = getInferenceRuleParams[ All, {{False, True|False}}, kbBrowserNB[]], ps, relPos, frLists, new, newForms, newRules, psChanged, changed = False},
		If[ MatchQ[ dr, {__, _}],
			new = Most[ dr];
			If[ MatchQ[ new, {String[ _String]..}],
				new = Join @@ Replace[
							new,
							String[ l_] :>
								With[ {aux = Cases[ $tmaEnv, _[ _, _, l, ___]]},
									If[ aux === {}, Sow[ {"Warning", "Formula (" <> l <> ") not found."}, "Message"]];
									aux
								],
							{1}
						]
			];
			new = DeleteCases[ Map[ With[ {tmp = simplifyFormulaByComputation[ #]}, If[ tmp === $Failed, #, tmp]]&, new], FML$[ _, True, ___]];
			If[ new === {},
				$Failed,
			(*else*)
				If[ TrueQ[ $autoGenerateRules],
					Block[ {$extractTagsForRewriting = True},
						frLists = Map[ Prepend[ trimFormulaForRewriting[ #], #]&, new]
					],
				(*else*)
					frLists = Map[ {#, {#}}&, new]
				];
				If[ TrueQ[ Last[ dr]],
					relPos = positionRelevantSits[ $TMAproofObject];
					If[ Length[ relPos] > 1,
						Scan[ Function[ p,
								ps = Extract[ $TMAproofObject, p];
								If[ id@ps =!= pid,
									{newForms, newRules} = getRelevantPair[ kb@ps, frLists];
									psChanged = False;
									If[ newForms =!= {},
										ps[[2]] = Join[ ps[[2]], newForms];
										psChanged = True
									];
									If[ newRules =!= {},
										ps = Replace[ ps, ((Rule|RuleDelayed)[ "rewriting", data_]) :> ("rewriting" -> addRulesToRW[ data, newRules]), {1}];
										psChanged = True;
									];
									If[ psChanged, $TMAproofObject = ReplacePart[ $TMAproofObject, p -> ps]; changed = True]
								]
							],
							relPos
						]
					]
				];
				
				{newForms, $rewriteRules} = getRelevantPair[ k, frLists];
				If[ changed || newForms =!= {} || $rewriteRules =!= {},
					addKnowledgeWhileProving[ new];
					makeANDNODE[ makePRFINFO[ name -> RRPaddKnowledge, used -> {}, generated -> {}],
						RRPToBeProved[ goal -> g, kb -> Join[ k, newForms], "newAssms" -> key /@ newForms, rest]
					],
				(*else*)
					$Failed
				]
			],
		(*else*)
			$Failed
		]
	]
]

getRelevantPair[ k_List, data_List] :=
	Module[ {f, r}, 
		f = Cases[ data, {fml_FML$, l_, rules___} /; !memberOfKB[ k, fml] :> {l, {rules}}];
		{f, r} = Switch[ f, {}, {{}, {}}, _, Transpose[ f]];
		{Join @@ f, DeleteCases[ r, {}]}
	]
	
inferenceRule[ RIremoveKnowledge] = 
this:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {rw = getOptionalComponent[ this, "rewriting"], sel = {}, delKeys},
		If[ ListQ[ $knowledgeActivated],
    		sel = Position[ k, FML$[ Alternatives@@$knowledgeActivated, __], {1}, Heads -> False]
    	];
		If[ sel === {},
			If[ Theorema`Provers`$allowInteraction =!= False, sel = selectFormulaDialog[ g, k, "removeKnowledge"]];
			If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ k],
				sel = {{sel}},
			(*else*)
				sel = $Failed
			]
		];
		If[ ListQ[ sel] && sel =!= {},
			delKeys = Alternatives@@Extract[ k, sel, key];
			rw = removeRulesFromRW[ rw, "keys" -> delKeys];
			makeANDNODE[ makePRFINFO[ name -> RIremoveKnowledge, used -> {}, generated -> {}],
				RRPToBeProved[ goal -> g, kb -> Delete[ k, sel], "rewriting" -> rw, rest]
			],
		(*else*)
			$Failed
		]
	]
]

inferenceRule[ RIhideKnowledge] = 
this:PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {sel = {}},
		If[ ListQ[ $knowledgeActivated],
    		sel = Position[ k, FML$[ Alternatives@@$knowledgeActivated, __], {1}, Heads -> False]
    	];
		If[ sel === {},
			If[ Theorema`Provers`$allowInteraction =!= False, sel = selectFormulaDialog[ g, k, "hideKnowledge"]];
			If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ k],
				sel = {{sel}},
			(*else*)
				sel = $Failed
			]
		];
		If[ ListQ[ sel] && sel =!= {},
			makeANDNODE[ makePRFINFO[ name -> RIhideKnowledge, used -> {}, generated -> {}],
				RRPToBeProved[ goal -> g, kb -> Delete[ k, sel], rest]
			],
		(*else*)
			$Failed
		]
	]
]


(* ::Section::Closed:: *)
(* Rule composition *)


(*
	Rule priorities:
	 0 - 19: termination + high priority
	20 - 24: medium priority, rewriting
	25 - 44: medium priority, goal-oriented
	45 - 64: medium priority, whole prfsit
	65 - 84: medium priority, kb-oriented
	85 - 89: low priority, goal-oriented
	90 - 94: low priority, kb-oriented
	95 - 99: low priority, whole prfsit + rewriting
*)

propositionalLogicRules = {"Propositional Logic Rules",
	{RRPgoalInKB, True, True, 1, "term"},
	{falseInKB, True, True, 1, "term"},
	{trueGoal, True, True, 1, "term"},
	{contradictionKB, True, True, 1, "term"},
	{RRPnotEquGoal, True, True, 40},
	{RRPandGoal, True, True, 85},
	{RRPandKB, True, False, 4},
	{RRPorGoal, True, True, 16},
	{RRPorKB, True, True, 91},
	{RRPimplGoalDirect, True, True, 10},
	{equivGoal, True, True, 41},
	{RRPequivKB, False, True, 90},
	{RRPsimpImplication, True, True, 70},
	{RRPmodusPonensTollens, True, True, 71},
	{RRPcontradiction, False, True, 99}
	};
	
miscRules = {"Miscellaneous Rules",
	{RRPsimplifyNegation, False, True, 50},
	{RRPexpandPiecewise, True, True, 95},
	{RIexpandSuch, True, True, 94}
	};
	
quantifierRules = {"Quantifier Rules",
	{RRPforallGoal, True, True, 19},
	{RRPsplitForallGoal, False, True, 88},
	{RRPexistsKB, True, True, 65},
	{RIforallGoalSeqLeft, True, True, 18}
	};
	
rewriteRules = {"Rules for Rewriting",
	{RRPgenerateRewriteRules, True, False, 2},
	{RRPgoalRewriting, True, True, 44},
	{RRPelementarySubstitution, True, True, 22},
	{RRPgroundSubstitution, True, True, 21},
	{RRPexpandDef, True, True, 97},
	{RRPimplicitDef, True, True, 97}
	};
	
interactiveRules = {"Interactive Rules",
	{RRPorGoalInteractive, False, True, 15},
	{RRPexchangeGoalKBInteractive, False, True, 100, "interactive"},
	{RRPsplitForallGoalInteractive, False, True, 100, "interactive"},
	{RRPinstantiateInteractive, True, True, 100, "interactive"},
	{RRPselectImplicationInteractive, True, True, 100, "interactive"},
	{RRPrewritePS, False, False, 100, "interactive"},
	{RRPintroduceFormula, True, True, 100, "interactive"},
	{RRPaddAbbreviation, True, True, 100, "interactive"}
	};
	
managementRules = {"Management Rules",
	{RRPaddKnowledge, False, False, 100, "interactive"},
	{RIremoveKnowledge, False, False, 100, "interactive"},
	{RIhideKnowledge, False, False, 100, "interactive"}
	};

registerRuleSet[ "RewriteInteractiveProver", rewriteInteractiveProver,
		{propositionalLogicRules, miscRules, quantifierRules, rewriteRules, interactiveRules, managementRules}
	]

(* ::Section:: *)
(* ProofScript Setup *)

Theorema`Provers`customInterpretationRules[ Hold[ rewriteInteractiveProver]] =
	{
		{Hold[ "Have"], Hold[ fml_Expression], post_} :> {{{"Apply", {"RRPintroduceFormula", fml, False, False}}}, post},
		{Hold[ "Split"], Hold[ "premise"], post_} :> {{{"Apply", {"RRPselectImplicationInteractive", False}}}, post},
		{Hold[ "Cases"], Hold[ fml_Expression], post_} :> {{{"Apply", {"RRPintroduceFormula", fml, True, False}}}, post},
		{Hold[ "Rewrite"], Hold[ fmls__, "by", rules__], post_Hold} :> {{{"Activate", {fmls}}, {"Apply", {"RRPrewritePS", rules}}}, post},
		{Hold[ "Rule"], Hold[], post_} :> {{{"Apply", {"RRPgoalRewriting"}}}, post}
	};

Theorema`Provers`customCommandSeparators[ Hold[ rewriteInteractiveProver]] =
	{"Have", "Split", "Cases", "Rule", "Rewrite"};

Theorema`Provers`customTranslationTable[ Hold[ rewriteInteractiveProver]] =
	{
		"allI" -> RRPforallGoal,
		"ccontr" -> RRPcontradiction,
		"conjE" -> RRPandKB,
		"conjI" -> RRPandGoal,
		"disjE" -> RRPorKB,
		"disjI" -> RRPorGoalInteractive,
		"esubs" -> RRPelementarySubstitution,
		"exchange" -> RRPexchangeGoalKBInteractive,
		"exE" -> RRPexistsKB,
		"forget" -> RIremoveKnowledge,
		"gsubs" -> RRPgroundSubstitution,
		"hide" -> RIhideKnowledge,
		"iffE" -> RRPequivKB,
		"iffI" -> equivGoal,
		"impI" -> RRPimplGoalDirect,
		"inst" -> RRPinstantiateInteractive,
		"know" -> RRPaddKnowledge,
		"mp" -> RRPmodusPonensTollens,
		"provePremise" -> RRPselectImplicationInteractive,
		"replImp" -> RIexpandSuch,
		"rewrite" -> RRPrewritePS,
		"rule" -> RRPgoalRewriting,
		"simpNeg" -> RRPsimplifyNegation,
		"splitAll" -> RRPsplitForallGoalInteractive,
		"splitPW" -> RRPexpandPiecewise,
		"unfold" -> RRPexpandDef,
		"unfoldImp" -> RRPimplicitDef
	};

End[]

EndPackage[]
